from __future__ import absolute_import
from __future__ import print_function

import os.path
import os
import time
import collections


SSG_PROJECT_NAME = "SCAP Security Guide Project"
SSG_BENCHMARK_LATEST_URI = "https://github.com/ComplianceAsCode/content/releases/latest"

SSG_REF_URIS = {
    'anssi': 'https://cyber.gouv.fr/sites/default/files/document/linux_configuration-en-v2.pdf',
    'bsi': 'https://www.bsi.bund.de/SharedDocs/Downloads/EN/BSI/Grundschutz/International/bsi_it_gs_comp_2022.pdf',
    'nist': 'http://nvlpubs.nist.gov/nistpubs/SpecialPublications/NIST.SP.800-53r4.pdf',
    'nist-csf': 'https://nvlpubs.nist.gov/nistpubs/CSWP/NIST.CSWP.04162018.pdf',
    'isa-62443-2013': 'https://www.isa.org/products/ansi-isa-62443-3-3-99-03-03-2013-security-for-indu',
    'isa-62443-2009': 'https://www.isa.org/products/isa-62443-2-1-2009-security-for-industrial-automat',
    'cobit5': 'https://www.isaca.org/resources/cobit',
    'cis-csc': 'https://www.cisecurity.org/controls/',
    'cjis': 'https://www.fbi.gov/file-repository/cjis-security-policy-v5_5_20160601-2-1.pdf',
    'cui': 'http://nvlpubs.nist.gov/nistpubs/SpecialPublications/NIST.SP.800-171.pdf',
    'dcid': 'not_officially_available',
    'disa': 'https://public.cyber.mil/stigs/cci/',
    'pcidss': 'https://www.pcisecuritystandards.org/documents/PCI_DSS_v3-2-1.pdf',
    'pcidss4': 'https://docs-prv.pcisecuritystandards.org/PCI%20DSS/Standard/PCI-DSS-v4_0.pdf',
    'ospp': 'https://www.niap-ccevs.org/Profile/PP.cfm',
    'hipaa': 'https://www.gpo.gov/fdsys/pkg/CFR-2007-title45-vol1/pdf/CFR-2007-title45-vol1-chapA-subchapC.pdf',
    'ism': 'https://www.cyber.gov.au/acsc/view-all-content/ism',
    'iso27001-2013': 'https://www.iso.org/contents/data/standard/05/45/54534.html',
    'nerc-cip': 'https://www.nerc.com/pa/Stand/AlignRep/One%20Stop%20Shop.xlsx',
    'stigid': 'https://public.cyber.mil/stigs/downloads/?_dl_facet_stigs=operating-systems%2Cunix-linux',
    'os-srg': 'https://public.cyber.mil/stigs/downloads/?_dl_facet_stigs=operating-systems%2Cgeneral-purpose-os',
    'app-srg': 'https://public.cyber.mil/stigs/downloads/?_dl_facet_stigs=application-servers',
    'app-srg-ctr': 'https://public.cyber.mil/stigs/downloads/?_dl_facet_stigs=container-platform',
    'stigref': 'https://public.cyber.mil/stigs/srg-stig-tools/',
}

product_directories = [
    'alinux2',
    'alinux3',
    'almalinux9',
    'anolis8',
    'anolis23',
    'al2023',
    'chromium',
    'debian11', 'debian12',
    'example',
    'eks',
    'fedora',
    'firefox',
    'macos1015',
    'kylinserver10',
    'ocp4',
    'rhcos4',
    'ol7', 'ol8', 'ol9', 'ol10',
    'openeuler2203',
    'opensuse',
    'openembedded',
    'rhel8', 'rhel9', 'rhel10',
    'rhv4',
    'sle12', 'sle15', 'slmicro5',
    'ubuntu1604', 'ubuntu1804', 'ubuntu2004', 'ubuntu2204', 'ubuntu2404'
]

JINJA_MACROS_DIRECTORY = os.path.abspath(os.path.join(os.path.dirname(os.path.dirname(
    __file__)), "shared", "macros"))

xml_version = """<?xml version="1.0" encoding="UTF-8"?>"""

datastream_namespace = "http://scap.nist.gov/schema/scap/source/1.2"
dc_namespace = "http://purl.org/dc/elements/1.1/"
ocil_namespace = "http://scap.nist.gov/schema/ocil/2.0"
cpe_language_namespace = "http://cpe.mitre.org/language/2.0"
cpe_dictionary_namespace = "http://cpe.mitre.org/dictionary/2.0"
oval_footer = "</oval_definitions>"
oval_namespace = "http://oval.mitre.org/XMLSchema/oval-definitions-5"
xlink_namespace = "http://www.w3.org/1999/xlink"
xhtml_namespace = "http://www.w3.org/1999/xhtml"
xsi_namespace = "http://www.w3.org/2001/XMLSchema-instance"
cat_namespace = "urn:oasis:names:tc:entity:xmlns:xml:catalog"
sce_namespace = "http://open-scap.org/page/SCE_xccdf_stream"
ocil_cs = "http://scap.nist.gov/schema/ocil/2"
bash_system = "urn:xccdf:fix:script:sh"
ansible_system = "urn:xccdf:fix:script:ansible"
ignition_system = "urn:xccdf:fix:script:ignition"
kubernetes_system = "urn:xccdf:fix:script:kubernetes"
blueprint_system = "urn:redhat:osbuild:blueprint"
puppet_system = "urn:xccdf:fix:script:puppet"
anaconda_system = "urn:redhat:anaconda:pre"
kickstart_system = "urn:xccdf:fix:script:kickstart"
bootc_system = "urn:xccdf:fix:script:bootc"
cce_uri = "https://ncp.nist.gov/cce"
stig_ns = "https://public.cyber.mil/stigs/srg-stig-tools/"
ccn_ns = "https://www.ccn-cert.cni.es/pdf/guias/series-ccn-stic/guias-de-acceso-publico-ccn-stic/6768-ccn-stic-610a22-perfilado-de-seguridad-red-hat-enterprise-linux-9-0/file.html"
cis_ns = "https://www.cisecurity.org/benchmark/red_hat_linux/"
hipaa_ns = "https://www.gpo.gov/fdsys/pkg/CFR-2007-title45-vol1/pdf/CFR-2007-title45-vol1-chapA-subchapC.pdf"
anssi_ns = "https://cyber.gouv.fr/sites/default/files/document/linux_configuration-en-v2.pdf"
ospp_ns = "https://www.niap-ccevs.org/Profile/PP.cfm"
pcidss4_ns = "https://docs-prv.pcisecuritystandards.org/PCI%20DSS/Standard/PCI-DSS-v4_0.pdf"
cui_ns = 'http://nvlpubs.nist.gov/nistpubs/SpecialPublications/NIST.SP.800-171.pdf'
stig_refs = 'https://public.cyber.mil/stigs/'
disa_cciuri = "https://public.cyber.mil/stigs/cci/"
ssg_version_uri = \
    "https://github.com/ComplianceAsCode/content/releases/latest"
OSCAP_VENDOR = "org.ssgproject"
OSCAP_DS_STRING = "xccdf_%s.content_benchmark_" % OSCAP_VENDOR
OSCAP_BENCHMARK = "xccdf_%s.content_benchmark_" % OSCAP_VENDOR
OSCAP_PROFILE = "xccdf_%s.content_profile_" % OSCAP_VENDOR
OSCAP_GROUP = "xccdf_%s.content_group_" % OSCAP_VENDOR
OSCAP_RULE = "xccdf_%s.content_rule_" % OSCAP_VENDOR
OSCAP_VALUE = "xccdf_%s.content_value_" % OSCAP_VENDOR
OSCAP_GROUP_PCIDSS = "xccdf_%s.content_group_pcidss-req" % OSCAP_VENDOR
OSCAP_GROUP_VAL = "xccdf_%s.content_group_values" % OSCAP_VENDOR
OSCAP_GROUP_NON_PCI = "xccdf_%s.content_group_non-pci-dss" % OSCAP_VENDOR
OSCAP_PATH = "oscap"
OSCAP_PROFILE_ALL_ID = "(all)"
XCCDF11_NS = "http://checklists.nist.gov/xccdf/1.1"
XCCDF12_NS = "http://checklists.nist.gov/xccdf/1.2"
min_ansible_version = "2.9"
ansible_version_requirement_pre_task_name = \
    "Verify Ansible meets SCAP-Security-Guide version requirements."
standard_profiles = ['standard', 'pci-dss', 'desktop', 'server']
xslt_ns = "http://www.w3.org/1999/XSL/Transform"
SCE_SYSTEM = "http://open-scap.org/page/SCE"


OVAL_SUB_NS = dict(
    ind="independent",
    unix="unix",
    linux="linux",
)


PREFIX_TO_NS = {
    "oval-def": oval_namespace,
    "oval": "http://oval.mitre.org/XMLSchema/oval-common-5",
    "dc": dc_namespace,
    "ds": datastream_namespace,
    "ocil": ocil_namespace,
    "xccdf-1.1": XCCDF11_NS,
    "xccdf-1.2": XCCDF12_NS,
    "html": xhtml_namespace,
    "xlink": xlink_namespace,
    "cpe-dict": cpe_dictionary_namespace,
    "cat": cat_namespace,
    "cpe-lang": cpe_language_namespace,
    "sce": sce_namespace,
}

FIX_TYPE_TO_SYSTEM = {
    "bash": bash_system,
    "ansible": ansible_system,
    "ignition": ignition_system,
    "kubernetes": kubernetes_system,
    "blueprint": blueprint_system,
    "puppet": puppet_system,
    "anaconda": anaconda_system,
    "kickstart": kickstart_system,
    "bootc": bootc_system,
}

for prefix, url_part in OVAL_SUB_NS.items():
    assert prefix not in PREFIX_TO_NS, \
        "Conflict between a namespace and OVAL sub-namespace '{prefix}'".format(prefix=prefix)
    PREFIX_TO_NS[prefix] = "{oval_ns}#{suffix}".format(oval_ns=PREFIX_TO_NS["oval-def"], suffix=url_part)


oval_header = (
    """
<oval_definitions
    xmlns="{0}"
    xmlns:oval="http://oval.mitre.org/XMLSchema/oval-common-5"
    xmlns:ind="{0}#independent"
    xmlns:unix="{0}#unix"
    xmlns:linux="{0}#linux"
    xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
    xsi:schemaLocation="http://oval.mitre.org/XMLSchema/oval-common-5 oval-common-schema.xsd
        {0} oval-definitions-schema.xsd
        {0}#independent independent-definitions-schema.xsd
        {0}#unix unix-definitions-schema.xsd
        {0}#linux linux-definitions-schema.xsd">"""
    .format(oval_namespace))

timestamp = time.strftime(
    "%Y-%m-%dT%H:%M:%S",
    time.gmtime(int(os.environ.get('SOURCE_DATE_EPOCH', time.time())))
)

PKG_MANAGER_TO_SYSTEM = {
    "yum": "rpm",
    "zypper": "rpm",
    "dnf": "rpm",
    "apt_get": "dpkg",
}

PKG_MANAGER_TO_CONFIG_FILE = {
    "yum": "/etc/yum.conf",
    "dnf": "/etc/dnf/dnf.conf",
    "zypper": "/etc/zypp/zypper.conf",
}

FULL_NAME_TO_PRODUCT_MAPPING = {
    "Alibaba Cloud Linux 2": "alinux2",
    "Alibaba Cloud Linux 3": "alinux3",
    "AlmaLinux OS 9": "almalinux9",
    "Anolis OS 8": "anolis8",
    "Anolis OS 23": "anolis23",
    "Amazon Linux 2023": "al2023",
    "Chromium": "chromium",
    "Debian 11": "debian11",
    "Debian 12": "debian12",
    "Example": "example",
    "Amazon Elastic Kubernetes Service": "eks",
    "Fedora": "fedora",
    "Firefox": "firefox",
    "Apple macOS 10.15": "macos1015",
    "Kylin Server 10": "kylinserver10",
    "Red Hat OpenShift Container Platform 4": "ocp4",
    "Red Hat Enterprise Linux CoreOS 4": "rhcos4",
    "Oracle Linux 7": "ol7",
    "Oracle Linux 8": "ol8",
    "Oracle Linux 9": "ol9",
    "Oracle Linux 10": "ol10",
    "openEuler 2203": "openeuler2203",
    "openSUSE": "opensuse",
    "Red Hat Enterprise Linux 8": "rhel8",
    "Red Hat Enterprise Linux 9": "rhel9",
    "Red Hat Enterprise Linux 10": "rhel10",
    "Red Hat Virtualization 4": "rhv4",
    "SUSE Linux Enterprise 12": "sle12",
    "SUSE Linux Enterprise 15": "sle15",
    "SUSE Linux Enterprise Micro 5": "slmicro5",
    "Ubuntu 16.04": "ubuntu1604",
    "Ubuntu 18.04": "ubuntu1804",
    "Ubuntu 20.04": "ubuntu2004",
    "Ubuntu 22.04": "ubuntu2204",
    "Ubuntu 24.04": "ubuntu2404",
    "OpenEmbedded": "openembedded",
    "Not Applicable": "example",
}


# see xccdf-addremediations.xslt <- shared_constants.xslt
# if you want to know how the map was constructed
REF_PREFIX_MAP = {
    "nist": "NIST-800-53",
    "cui": "NIST-800-171",
    "pcidss": "PCI-DSS",
    "pcidss4": "PCI-DSSv4",
    "cjis": "CJIS",
    "stigid": "DISA-STIG",
}

Reference = collections.namedtuple("Reference", ("id", "name", "url", "regex_with_groups"))

REFERENCES = dict(
    anssi=Reference(
        id="anssi", name="ANSSI", url=anssi_ns,
        regex_with_groups=r"R(\d+)"),
    ccn=Reference(
        id="ccn", name="CCN", url="",
        regex_with_groups=r"A\.(\d+)\.SEC-(\w+)(\d+)"),
    cis=Reference(
        id="cis", name="CIS", url=cis_ns,
        regex_with_groups=r"(\d+)\.(\d+)(?:\.(\w+)(?:\.(\w+)(?:\.(\w+))?)?)?"),
    cui=Reference(
        id="cui", name=REF_PREFIX_MAP["cui"], url=cui_ns,
        regex_with_groups=r"(\d+)(?:\.(\w+)(?:\.(\w+)(?:\.(\w+))?)?)?"),
    nist=Reference(
        id="nist", name=REF_PREFIX_MAP["nist"], url="",
        regex_with_groups=r".*-(\d+)(?:\((\d+)\))?"),
    ospp=Reference(
        id="ospp", name="OSPP", url=SSG_REF_URIS["ospp"],
        regex_with_groups=r"(\w+)(?:\.(\d+)(?:\.([^\.]+)(?:\.([^\.]+))?)?)?"),
    pcidss=Reference(
        id="pcidss", name=REF_PREFIX_MAP["pcidss"], url="",
        regex_with_groups=r"Req-(\d+)(?:\.(\w+)(?:\.(\w+)(?:\.(\w+))?)?)?"),
    pcidss4=Reference(
        id="pcidss4", name=REF_PREFIX_MAP["pcidss4"], url="",
        regex_with_groups=r"(\d+)(?:\.(\w+)(?:\.(\w+)(?:\.(\w+))?)?)?"),
    srg=Reference(
        id="srg", name="SRG", url="",
        regex_with_groups=r"(SRG-OS-\d+-GPOS-\d+)"
    )
)


MULTI_PLATFORM_LIST = ["rhel", "fedora", "rhv", "debian", "ubuntu",
                       "openeuler", "kylinserver",
                       "opensuse", "sle", "ol", "ocp", "rhcos",
                       "example", "eks", "alinux", "anolis", "openembedded", "al",
                       "slmicro", "almalinux"]

MULTI_PLATFORM_MAPPING = {
    "multi_platform_alinux": ["alinux2", "alinux3"],
    "multi_platform_almalinux": ["almalinux9"],
    "multi_platform_anolis": ["anolis8", "anolis23"],
    "multi_platform_debian": ["debian11", "debian12"],
    "multi_platform_example": ["example"],
    "multi_platform_eks": ["eks"],
    "multi_platform_fedora": ["fedora"],
    "multi_platform_kylinserver": ["kylinserver10"],
    "multi_platform_openeuler": ["openeuler2203"],
    "multi_platform_opensuse": ["opensuse"],
    "multi_platform_ol": ["ol7", "ol8", "ol9", "ol10"],
    "multi_platform_ocp": ["ocp4"],
    "multi_platform_rhcos": ["rhcos4"],
    "multi_platform_rhel": ["rhel8", "rhel9", "rhel10"],
    "multi_platform_rhv": ["rhv4"],
    "multi_platform_sle": ["sle12", "sle15"],
    "multi_platform_slmicro": ["slmicro5"],
    "multi_platform_ubuntu": ["ubuntu1604", "ubuntu1804", "ubuntu2004",
                              "ubuntu2204", "ubuntu2404"],
    "multi_platform_openembedded": ["openembedded"],
    "multi_platform_al": ["al2023"],
}

RHEL_CENTOS_CPE_MAPPING = {
    "cpe:/o:redhat:enterprise_linux:8": "cpe:/o:centos:centos:8",
    "cpe:/o:redhat:enterprise_linux:9": "cpe:/o:centos:centos:9",
    "cpe:/o:redhat:enterprise_linux:10": "cpe:/o:centos:centos:10",
}

CENTOS_NOTICE = \
    "<div xmlns=\"http://www.w3.org/1999/xhtml\">\n" \
    "<p>This benchmark is a direct port of a <i>SCAP Security Guide </i> " \
    "benchmark developed for <i>Red Hat Enterprise Linux</i>. It has been " \
    "modified through an automated process to remove specific dependencies " \
    "on <i>Red Hat Enterprise Linux</i> and to function with <i>CentOS</i>. " \
    "The result is a generally useful <i>SCAP Security Guide</i> benchmark " \
    "with the following caveats:</p>\n" \
    "<ul>\n" \
    "<li><i>CentOS</i> is not an exact copy of " \
    "<i>Red Hat Enterprise Linux</i>. There may be configuration differences " \
    "that produce false positives and/or false negatives. If this occurs " \
    "please file a bug report.</li>\n" \
    "\n" \
    "<li><i>CentOS</i> has its own build system, compiler options, patchsets, " \
    "and is a community supported, non-commercial operating system. " \
    "<i>CentOS</i> does not inherit " \
    "certifications or evaluations from <i>Red Hat Enterprise Linux</i>. As " \
    "such, some configuration rules (such as those requiring " \
    "<i>FIPS 140-2</i> encryption) will continue to fail on <i>CentOS</i>.</li>\n" \
    "</ul>\n" \
    "\n" \
    "<p>Members of the <i>CentOS</i> community are invited to participate in " \
    "<a href=\"http://open-scap.org\">OpenSCAP</a> and " \
    "<a href=\"https://github.com/ComplianceAsCode/content\">" \
    "SCAP Security Guide</a> development. Bug reports and patches " \
    "can be sent to GitHub: " \
    "<a href=\"https://github.com/ComplianceAsCode/content\">" \
    "https://github.com/ComplianceAsCode/content</a>. " \
    "The mailing list is at " \
    "<a href=\"https://fedorahosted.org/mailman/listinfo/scap-security-guide\">" \
    "https://fedorahosted.org/mailman/listinfo/scap-security-guide</a>" \
    ".</p>" \
    "</div>"

XCCDF_REFINABLE_PROPERTIES = ["weight", "severity", "role", "selector"]

OVAL_TO_XCCDF_DATATYPE_CONSTRAINTS = {
    'int': 'number',
    'float': 'number',
    'boolean': 'boolean',
    'string': 'string',
    'evr_string': 'string',
    'version': 'string',
    'ios_version': 'string',
    'fileset_revision': 'string',
    'binary': 'string'
}

OVALTAG_TO_ABBREV = {
    'definition': 'def',
    'criteria': 'crit',
    'test': 'tst',
    'object': 'obj',
    'state': 'ste',
    'variable': 'var',
}

OCILTAG_TO_ABBREV = {
    'questionnaire': 'questionnaire',
    'action': 'testaction',
    'question': 'question',
    'artifact': 'artifact',
    'variable': 'variable',
}

OVALREFATTR_TO_TAG = {
    "definition_ref": "definition",
    "test_ref": "test",
    "object_ref": "object",
    "state_ref": "state",
    "var_ref": "variable",
}

OCILREFATTR_TO_TAG = {
    "question_ref": "question",
}

# Default platform to package mapping
XCCDF_PLATFORM_TO_PACKAGE = {
  "grub2": "grub2-common",
  "login_defs": "login",
  "sssd": "sssd-common",
  "zipl": "s390utils-base",
  "sssd-ldap": None,  # Force package check wrapping skip
  "uefi": None,
  "non-uefi": None,
  "not_s390x_arch": None,
  "s390x_arch": None,
  "not_aarch64_arch": None,
  "aarch64_arch": None,
  "ovirt": None,
  "no_ovirt": None,
}

# _version_name_map = {
MAKEFILE_ID_TO_PRODUCT_MAP = {
    'alinux': 'Alibaba Cloud Linux',
    'almalinux': 'AlmaLinux OS',
    'anolis': 'Anolis OS',
    'chromium': 'Google Chromium Browser',
    'fedora': 'Fedora',
    'firefox': 'Mozilla Firefox',
    'macos': 'Apple macOS',
    'kylinserver': 'Kylin Server',
    'rhel': 'Red Hat Enterprise Linux',
    'rhv': 'Red Hat Virtualization',
    'debian': 'Debian',
    'ubuntu': 'Ubuntu',
    'eap': 'JBoss Enterprise Application Platform',
    'fuse': 'JBoss Fuse',
    'openeuler': 'openEuler',
    'opensuse': 'openSUSE',
    'sle': 'SUSE Linux Enterprise',
    'slmicro': 'SUSE Linux Enterprise Micro',
    'example': 'Example',
    'ol': 'Oracle Linux',
    'ocp': 'Red Hat OpenShift Container Platform',
    'rhcos': 'Red Hat Enterprise Linux CoreOS',
    'eks': 'Amazon Elastic Kubernetes Service',
    'al': 'Amazon Linux',
    'openembedded': 'OpenEmbedded',
}

# References that can not be used with product-qualifiers
GLOBAL_REFERENCES = ("srg", "disa", "cis-csc",)

# Application constants
DEFAULT_DCONF_GDM_DIR = 'gdm.d'
DEFAULT_AIDE_CONF_PATH = '/etc/aide.conf'
DEFAULT_AIDE_BIN_PATH = '/usr/sbin/aide'
DEFAULT_FAILLOCK_PATH = '/var/run/faillock'
DEFAULT_SSH_DISTRIBUTED_CONFIG = 'false'
DEFAULT_PRODUCT = 'example'
DEFAULT_CHRONY_CONF_PATH = '/etc/chrony.conf'
DEFAULT_CHRONY_D_PATH = '/etc/chrony.d/'
DEFAULT_AUDISP_CONF_PATH = '/etc/audit'
DEFAULT_SYSCTL_REMEDIATE_DROP_IN_FILE = 'false'
DEFAULT_BOOTABLE_CONTAINERS_SUPPORTED = 'false'


# Constants for OVAL object model
STR_TO_BOOL = {
    "false": False,
    "False": False,
    "true": True,
    "True": True,
}

BOOL_TO_STR = {True: "true", False: "false"}


class OvalNamespaces:
    oval = "http://oval.mitre.org/XMLSchema/oval-common-5"
    definition = oval_namespace
    independent = "http://oval.mitre.org/XMLSchema/oval-definitions-5#independent"
    linux = "http://oval.mitre.org/XMLSchema/oval-definitions-5#linux"


OVAL_NAMESPACES = OvalNamespaces()

DERIVATIVES_PRODUCT_MAPPING = {
    "centos8": "rhel8",
    "cs9": "rhel9",
    "cs10": "rhel10",
}

BENCHMARKS = {
    "apple_os",
    "applications",
    "linux_os/guide",
    "products/chromium/guide",
    "products/firefox/guide",
}


SSG_IDENT_URIS = {
    'cce': cce_uri
}



from ssg.utils import parse_template_boolean_value

def preprocess(data, lang):
    data["oval_extend_definitions"] = data.get("oval_extend_definitions", [])
    data["escape_text"] = parse_template_boolean_value(data,
                                                     parameter="escape_text",
                                                     default_value=True)
    return data



from .sub import command_sub
from .stats import command_stats
from .most_used_rules import command_most_used_rules
from .most_used_components import command_most_used_components
from .common import generate_output



import pytest

from ssg import boolean_expression


@pytest.fixture
def algebra():
    return boolean_expression.Algebra(
        symbol_cls=boolean_expression.Symbol, function_cls=boolean_expression.Function)


@pytest.fixture
def expression1(algebra):
    return algebra.parse(u'(oranges | banana) and not ~apple + !pie', simplify=True)


def test_dynamic_algebra():
    alg = boolean_expression.Algebra(
        symbol_cls=boolean_expression.Symbol, function_cls=boolean_expression.Function)
    exp = alg.parse('not banana and not apple or anything')
    assert str(exp) == '(~banana&~apple)|anything'
    assert str(exp.simplify()) == 'anything|(~apple&~banana)'


def test_compare_expressions(algebra):
    exp1 = algebra.parse('banana1 and apple1')
    exp2 = algebra.parse('apple1 and banana1')
    assert exp1 == exp2


def test_cnf(algebra, expression1):
    assert str(algebra.cnf(expression1)) == '(apple|~pie)&(banana|oranges|~pie)'


def test_dnf(algebra, expression1):
    assert str(algebra.dnf(expression1)) == '(apple&banana)|(apple&oranges)|~pie'


def test_underscores_and_dashes_in_name(algebra):
    exp = algebra.parse(u'not_s390x_arch and dashed-name')
    assert exp(**{'not_s390x_arch': True, 'dashed-name': True})


def test_expression_with_parameter(algebra):
    exp = algebra.parse(u'package[test] and os_release[rhel]')
    assert exp(**{"package[test]": True, "os_release[rhel]": True})


def test_evaluate_simple_boolean_ops(algebra):
    exp = algebra.parse(u'(oranges | banana) and not not apple or !pie')
    assert exp(**{'oranges': True, 'apple': True, 'pie': True})
    assert not exp(**{'oranges': True, 'apple': False, 'pie': True})


def test_args_non_versioned(algebra):
    exp = algebra.parse(u'oranges[aaa] | oranges[bbb]')
    assert exp(**{'oranges[aaa]': True, 'oranges[bbb]': True})
    assert not exp(**{'oranges[zzz]': True})
    assert not exp(**{'oranges': True})


def test_args_and_version(algebra):
    exp = algebra.parse(u'oranges[aaa] & oranges[bbb]>=2')
    assert exp(**{'oranges[aaa]': True, 'oranges[bbb]': '2'})
    assert not exp(**{'oranges[aaa]': True, 'oranges[bbb]': '1'})
    assert not exp(**{'oranges[zzz]': True})
    assert not exp(**{'oranges': True})
    assert not exp(**{'oranges[bbb,zzz]': True})


def test_evaluate_simple_version_ops(algebra):
    exp = algebra.parse(u'oranges==2')
    assert exp(**{'oranges': '2'})
    assert exp(**{'oranges': '2.0'})
    assert exp(**{'oranges': '2.0.0'})
    assert not exp(**{'oranges': '2.0.1'})
    assert not exp(**{'oranges': '3.0'})
    assert not exp(**{'oranges': True})
    assert not exp(**{'oranges': 2})
    assert not exp(**{'oranges': 2.0})


def test_evaluate_advanced_version_ops(algebra):
    exp = algebra.parse(u'oranges>=1.0,<3.0 and oranges!=2.6')
    assert exp(**{'oranges': '2'})
    assert exp(**{'oranges': '2.9'})
    assert exp(**{'oranges': '2.0.1'})
    assert exp(**{'oranges': '2.9.0-1'})
    assert not exp(**{'oranges': '3.0'})
    assert not exp(**{'oranges': '0.9.999'})
    assert not exp(**{'oranges': '2.6.0'})





import pytest

from ssg_test_suite import common


def test_simple_match():
    scenario_platforms = ["Red Hat Enterprise Linux 10"]
    benchmark_cpes = {"cpe:/o:redhat:enterprise_linux:10"}
    assert common.matches_platform(scenario_platforms, benchmark_cpes) is True


def test_simple_no_match():
    scenario_platforms = ["Red Hat Enterprise Linux 10"]
    benchmark_cpes = {"cpe:/o:redhat:enterprise_linux:9"}
    assert common.matches_platform(scenario_platforms, benchmark_cpes) is False


def test_multi_platform_all():
    scenario_platforms = ["multi_platform_all"]
    benchmark_cpes = {"cpe:/o:redhat:enterprise_linux:10"}
    assert common.matches_platform(scenario_platforms, benchmark_cpes) is True


def test_multi_platform_match():
    scenario_platforms = ["multi_platform_rhel"]
    benchmark_cpes = {"cpe:/o:redhat:enterprise_linux:10"}
    assert common.matches_platform(scenario_platforms, benchmark_cpes) is True


def test_multi_platform_no_match():
    scenario_platforms = ["multi_platform_fedora"]
    benchmark_cpes = {"cpe:/o:redhat:enterprise_linux:10"}
    assert common.matches_platform(scenario_platforms, benchmark_cpes) is False


def test_list_simple_match_first():
    scenario_platforms = ["Red Hat Enterprise Linux 9",
                          "Red Hat Enterprise Linux 10"]
    benchmark_cpes = {"cpe:/o:redhat:enterprise_linux:9"}
    assert common.matches_platform(scenario_platforms, benchmark_cpes) is True


def test_list_simple_match_second():
    scenario_platforms = ["Red Hat Enterprise Linux 9",
                          "Red Hat Enterprise Linux 10"]
    benchmark_cpes = {"cpe:/o:redhat:enterprise_linux:10"}
    assert common.matches_platform(scenario_platforms, benchmark_cpes) is True


def test_list_simple_no_match():
    scenario_platforms = ["Red Hat Enterprise Linux 8",
                          "Red Hat Enterprise Linux 9"]
    benchmark_cpes = {"cpe:/o:redhat:enterprise_linux:10"}
    assert common.matches_platform(scenario_platforms, benchmark_cpes) is False


def test_list_multi_platform_match_first():
    scenario_platforms = ["multi_platform_rhel", "multi_platform_debian"]
    benchmark_cpes = {"cpe:/o:redhat:enterprise_linux:8"}
    assert common.matches_platform(scenario_platforms, benchmark_cpes) is True


def test_list_multi_platform_match_second():
    scenario_platforms = ["multi_platform_rhel", "multi_platform_debian"]
    benchmark_cpes = {"cpe:/o:debian:debian_linux:11"}
    assert common.matches_platform(scenario_platforms, benchmark_cpes) is True


def test_list_multi_platform_no_match():
    scenario_platforms = ["multi_platform_debian", "multi_platform_ubuntu"]
    benchmark_cpes = {"cpe:/o:redhat:enterprise_linux:8"}
    assert common.matches_platform(scenario_platforms, benchmark_cpes) is False


def test_list_combined_match():
    scenario_platforms = ["multi_platform_debian", "multi_platform_ubuntu",
                          "Red Hat Enterprise Linux 8"]
    benchmark_cpes = {"cpe:/o:redhat:enterprise_linux:8"}
    assert common.matches_platform(scenario_platforms, benchmark_cpes) is True


def test_list_combined_match_2():
    scenario_platforms = ["Debian 11", "multi_platform_ubuntu", "openSUSE",
                          "Red Hat Enterprise Linux 8"]
    benchmark_cpes = {"cpe:/o:debian:debian_linux:11"}
    assert common.matches_platform(scenario_platforms, benchmark_cpes) is True


def test_list_combined_no_match():
    scenario_platforms = ["multi_platform_ubuntu", "multi_platform_fedora",
                          "Red Hat Enterprise Linux 8", "openSUSE"]
    benchmark_cpes = {"cpe:/o:debian:debian_linux:11"}
    assert common.matches_platform(scenario_platforms, benchmark_cpes) is False


def test_simple_multiple_unrelated_benchmark_cpes():
    scenario_platforms = ["Red Hat Enterprise Linux 10"]
    benchmark_cpes = {"cpe:/o:redhat:enterprise_linux:9",
                      "cpe:/o:redhat:enterprise_linux:10"}
    assert common.matches_platform(scenario_platforms, benchmark_cpes) is True


def test_simple_multiple_bogus_benchmark_cpes():
    scenario_platforms = ["Red Hat Enterprise Linux 10"]
    benchmark_cpes = {"cpe:/o:abcdef:ghijklm:42"
                      "cpe:/o:zzzzz:xxxx:77",
                      "cpe:/o:redhat:enterprise_linux:10"}
    assert common.matches_platform(scenario_platforms, benchmark_cpes) is True

def test_simple_multiple_bogus_benchmark_cpes_no_match():
    scenario_platforms = ["Fedora"]
    benchmark_cpes = {"cpe:/o:abcdef:ghijklm:42"
                      "cpe:/o:zzzzz:xxxx:77",
                      "cpe:/o:redhat:enterprise_linux:10"}
    assert common.matches_platform(scenario_platforms, benchmark_cpes) is False


def test_multiple_multiple_bogus_benchmark_cpes_no_match():
    scenario_platforms = ["Fedora", "openSUSE"]
    benchmark_cpes = {"cpe:/o:abcdef:ghijklm:42"
                      "cpe:/o:zzzzz:xxxx:77",
                      "cpe:/o:redhat:enterprise_linux:10"}
    assert common.matches_platform(scenario_platforms, benchmark_cpes) is False


def test_typo():
    scenario_platforms = ["Rrd Hat Enterprise Linux 10"]
    benchmark_cpes = {"cpe:/o:redhat:enterprise_linux:10"}
    with pytest.raises(ValueError):
        common.matches_platform(scenario_platforms, benchmark_cpes)


def test_wrong_multi_platform():
    scenario_platforms = ["multi_platform_fidorka"]
    benchmark_cpes = {"cpe:/o:fedoraproject:fedora:30"}
    with pytest.raises(ValueError):
        common.matches_platform(scenario_platforms, benchmark_cpes)



"""
Common functions for Boolean Expressions
"""

from ssg.ext.boolean import boolean
from ssg import requirement_specs

# We don't support ~= to avoid confusion with boolean operator NOT (~)
SPEC_SYMBOLS = ['<', '>', '=', '!', ',', '[', ']']
VERSION_SYMBOLS = ['.', '-', '_', ":"]


class Function(boolean.Function):
    """
    Base class for boolean functions.

    This class should be subclassed and passed to the `Algebra` as `function_cls`
    to enrich expression elements with domain-specific methods.
    """
    def is_and(self):
        return isinstance(self, boolean.AND)

    def is_or(self):
        return isinstance(self, boolean.OR)

    def is_not(self):
        return isinstance(self, boolean.NOT)

    def as_id(self):
        """
        Generate a string representation of the boolean expression.

        This method constructs a unique identifier for the boolean expression by recursively
        calling `as_id` on its arguments and combining them with the appropriate boolean operator.

        Returns:
            str: A string representing the boolean expression. If the expression is a negation, it
                 returns 'not_' followed by the identifier of the negated argument. If the
                 expression is a conjunction or disjunction, it returns the identifiers of the
                 arguments joined by '_and_' or '_or_'.
        """
        if self.is_not():
            return 'not_{0}'.format(self.args[0].as_id())
        op = 'unknown_bool_op'
        if self.is_and():
            op = 'and'
        if self.is_or():
            op = 'or'
        return '_{0}_'.format(op).join([arg.as_id() for arg in self.args])


class Symbol(boolean.Symbol):
    """
    Symbol class represents a boolean symbol with domain-specific methods.

    This class should be subclassed and passed to the `Algebra` as `symbol_cls`
    to enrich expression elements with domain-specific methods.

    Attributes:
        requirement (requirement_specs.Requirement): The requirement object associated with the symbol.
        obj (requirement_specs.Requirement): Alias for the requirement attribute.
    """
    def __init__(self, obj):
        super(Symbol, self).__init__(obj)
        self.requirement = requirement_specs.Requirement(obj)
        self.obj = self.requirement

    def __call__(self, **kwargs):
        full_name = self.name
        if self.arg:
            full_name += '[' + self.arg + ']'
        val = kwargs.get(full_name, False)
        if self.requirement.has_version_specs():
            if type(val) is str:
                return val in self.requirement
            return False
        return bool(val)

    def __hash__(self):
        return hash(self.as_id())

    def __eq__(self, other):
        return hash(self) == hash(other)

    def __lt__(self, other):
        return self.as_id() < other.as_id()

    def as_id(self):
        id_str = self.name
        if self.arg:
            id_str += '_' + self.arg
        if self.requirement.has_version_specs():
            id_str += '_' + self.requirement.ver_specs.oval_id
        return id_str

    def as_dict(self):
        res = {
            'id': self.as_id(),
            'name': self.name,
            'arg': self.arg,
            'ver_specs': [],
            'ver_specs_id': '',
            'ver_specs_cpe': '',
            'ver_specs_title': '',
        }

        if self.requirement.has_version_specs():
            for ver_spec in sorted(self.requirement.ver_specs):
                res['ver_specs'].append({
                    'id': ver_spec.oval_id,
                    'op': ver_spec.op,
                    'ver': ver_spec.ver,
                    'evr_op': ver_spec.evr_op,
                    'evr_ver': ver_spec.evr_ver,
                    'ev_ver': ver_spec.ev_ver
                })
            res['ver_specs_id'] = self.requirement.ver_specs.oval_id
            res['ver_specs_cpe'] = self.requirement.ver_specs.cpe_id
            res['ver_specs_title'] = self.requirement.ver_specs.title

        return res

    def has_version_specs(self):
        return self.requirement.has_version_specs()

    @property
    def arg(self):
        return self.requirement.arg or ''

    @property
    def name(self):
        return self.requirement.name

    @staticmethod
    def is_parametrized(name):
        return requirement_specs.Requirement.is_parametrized(name)

    @staticmethod
    def get_base_of_parametrized_name(name):
        return requirement_specs.Requirement.get_base_for_parametrized(name)


class Algebra(boolean.BooleanAlgebra):
    """
    Base class for boolean algebra.

    The Algebra class parses and evaluates boolean expressions, where operators
    can be any combination of "~, &, |, !, *, +, not, and, or" and variable symbols
    can contain version specifiers as described in PEP440 and PEP508.
    - No white space is allowed inside specifier expressions.
    - The ~= specifier operator is not supported.

    Limitations:
    - no white space is allowed inside specifier expressions;
    - ~= specifier operator is not supported.

    Example:
        "(oranges>=2.0.8,<=5 | fried[banana]) and !pie[apple]"

    Attributes:
        symbol_cls (class): The class used for symbols in the boolean expressions.
        function_cls (class): The class used for functions in the boolean expressions.
    """
    def __init__(self, symbol_cls, function_cls):
        not_cls = type('FunctionNOT', (function_cls, boolean.NOT), {})
        and_cls = type('FunctionAND', (function_cls, boolean.AND), {})
        or_cls = type('FunctionOR', (function_cls, boolean.OR), {})
        super(Algebra, self).__init__(allowed_in_token=VERSION_SYMBOLS+SPEC_SYMBOLS,
                                      Symbol_class=symbol_cls,
                                      NOT_class=not_cls, AND_class=and_cls, OR_class=or_cls)



#!/usr/bin/python3

from __future__ import print_function

import argparse

import ssg.environment
import ssg.templates


def parse_args():
    p = argparse.ArgumentParser()
    p.add_argument(
        "--build-config-yaml", required=True,
        help="YAML file with information about the build configuration. "
        "e.g.: ~/scap-security-guide/build/build_config.yml"
    )
    p.add_argument(
        "--product-yaml", required=True,
        help="YAML file with information about the product we are building. "
        "e.g.: ~/scap-security-guide/rhel9/product.yml"
    )
    p.add_argument(
        "--resolved-rules-dir", required=True,
        help="Directory with <rule-id>.yml resolved rule YAMLs. "
        "e.g.: ~/scap-security-guide/build/rhel9/rules"
    )
    p.add_argument(
        "--templates-dir", required=True,
        help="Path to directory which contains content templates. "
        "e.g.: ~/scap-security-guide/shared/templates"
    )
    p.add_argument(
        "--checks-dir", required=True,
        help="Path to which OVAL checks will be generated. "
        "e.g.: ~/scap-security-guide/build/rhel9/checks"
    )
    p.add_argument(
        "--platforms-dir", required=True,
        help="Path to directory which contains prebuilt platforms. "
             "e.g.: ~/scap-security-guide/build/rhel9/platforms"
    )
    p.add_argument(
        "--cpe-items-dir", required=True,
        help="Path to directory which contains compiled CPE items. "
             "e.g.: ~/scap-security-guide/build/rhel9/cpe_items"
    )
    p.add_argument(
        "--remediations-dir", required=True,
        help="Path to which remediations will be generated. "
        "e.g.: ~/scap-security-guide/build/rhel9/fixes_from_templates"
    )
    args = p.parse_args()
    return args


if __name__ == "__main__":
    args = parse_args()

    env_yaml = ssg.environment.open_environment(
        args.build_config_yaml, args.product_yaml)
    builder = ssg.templates.Builder(
        env_yaml, args.resolved_rules_dir, args.templates_dir,
        args.remediations_dir, args.checks_dir, args.platforms_dir, args.cpe_items_dir)
    builder.build()



#!/usr/bin/python3

import argparse
import os
import glob
import sys

import ssg.constants

SSG_ROOT = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))
VALID_FIELDS = ['check', 'packages', 'platform', 'profiles', 'remediation', 'templates',
                'variables']
VALID_STATES = ['pass', 'fail', 'notapplicable']

VALID_PLATFORMS = (list(ssg.constants.FULL_NAME_TO_PRODUCT_MAPPING.keys())
                   + list(ssg.constants.MULTI_PLATFORM_MAPPING.keys())
                   + ['multi_platform_all'])


def _parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser()
    parser.add_argument("-r", "--root", required=False, default=SSG_ROOT,
                        help="Root directory of the project")
    return parser.parse_args()


def get_files(root: str):
    result = glob.glob("linux_os/**/tests/*.sh", recursive=True, root_dir=root)
    return result


def _test_filename_valid(test_file: str) -> bool:
    filename = os.path.basename(test_file)
    end_state = filename.split('.')
    if len(end_state) == 3 and end_state[1] not in VALID_STATES:
        print(f"Invalid expected state '{end_state[1]}' in {test_file}", file=sys.stderr)
        return False
    return True


def _validate_platform(param_value, test_file):
    for platform in param_value.split(","):
        if platform.strip() not in VALID_PLATFORMS:
            print(f"Invalid platform '{platform}' in {test_file}", file=sys.stderr)
            return False
    return True


def _has_invalid_param(root: str, test_file: str) -> bool:
    full_path = os.path.join(root, test_file)
    has_no_errors = True
    with open(full_path, "r") as f:
        for line in f:
            if not line.startswith("#"):
                break
            line = line.removeprefix('#')
            line = line.strip()
            parts = line.split('=')
            if len(parts) != 2:
                continue
            param_name = parts[0].strip()
            param_value = parts[1].strip()
            if param_name == 'platform':
                has_no_errors = _validate_platform(param_value, test_file)
            if param_name not in VALID_FIELDS:
                print(f"Invalid field '{param_name}' in {test_file}", file=sys.stderr)
                has_no_errors = False
    return has_no_errors


def main() -> int:
    args = _parse_args()
    test_files = get_files(args.root)
    return_value = 0
    for test_file in test_files:
        if not _test_filename_valid(test_file):
            return_value = 1
        if not _has_invalid_param(args.root, test_file):
            return_value = 1
    return return_value


if __name__ == "__main__":
    raise SystemExit(main())



#!/usr/bin/python3

import argparse
import xml.etree.ElementTree as ET
import yaml

from ssg.constants import datastream_namespace, XCCDF12_NS


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("datastream", help="SCAP source data stream")
    parser.add_argument("product_yaml", help="Resolved product YAML")
    return parser.parse_args()


def get_references_in_benchmark(ds_path):
    references = {}
    root = ET.parse(ds_path).getroot()
    benchmark_xpath = "./{%s}component/{%s}Benchmark" % (datastream_namespace, XCCDF12_NS)
    benchmark_el = root.find(benchmark_xpath)
    reference_xpath = "{%s}reference" % XCCDF12_NS
    for reference_el in benchmark_el.findall(reference_xpath):
        href = reference_el.get("href")
        title = reference_el.text
        references[title] = href
    return references


def get_references_in_product_yaml(product_yaml_path):
    with open(product_yaml_path) as product_yaml_fd:
        product_yaml = yaml.safe_load(product_yaml_fd)
    product_yaml_references = product_yaml["reference_uris"]
    return product_yaml_references


def main():
    args = parse_args()
    references_in_benchmark = get_references_in_benchmark(args.datastream)
    references_in_product_yaml = get_references_in_product_yaml(args.product_yaml)
    assert references_in_benchmark == references_in_product_yaml


if __name__ == "__main__":
    main()



from ssg.utils import parse_template_boolean_value
from ssg.utils import ensure_file_paths_and_file_regexes_are_correctly_defined


def preprocess(data, lang):
    ensure_file_paths_and_file_regexes_are_correctly_defined(data)

    data["recursive"] = parse_template_boolean_value(data,
                                                     parameter="recursive",
                                                     default_value=False)

    try:
        int(data["gid_or_name"])
        data["group_represented_with_gid"] = True
    except ValueError:
        data["group_represented_with_gid"] = False

    if lang == "oval":
        data["fileid"] = data["_rule_id"].replace("file_groupowner", "")
    return data



from ssg.utils import parse_template_boolean_value

import ssg.utils
import re


def _mount_option(data, lang):
    if lang == "oval":
        data["pointid"] = ssg.utils.escape_id(data["mountpoint"])
        data["pointregex"] = ssg.utils.escape_regex(data["mountpoint"])
    else:
        data["mountoption"] = re.sub(" ", ",", data["mountoption"])
    return data


def preprocess(data, lang):
    data["mount_has_to_exist"] = parse_template_boolean_value(data,
                                                              parameter="mount_has_to_exist",
                                                              default_value=True)
    if lang == "oval":
        data["mountoptionid"] = data["mountoption"].split()[0]
    return _mount_option(data, lang)



from __future__ import absolute_import
from __future__ import print_function

import sys
import os
import os.path
import re
from collections import defaultdict, namedtuple, OrderedDict

import ssg.yaml
import ssg.build_yaml
from . import rules
from . import utils

from . import constants
from .jinja import process_file_with_macros as jinja_process_file

from .xml import ElementTree
from .constants import XCCDF12_NS

REMEDIATION_TO_EXT_MAP = {
    'anaconda': '.anaconda',
    'ansible': '.yml',
    'bash': '.sh',
    'puppet': '.pp',
    'ignition': '.yml',
    'kubernetes': '.yml',
    'blueprint': '.toml',
    'kickstart': '.cfg',
    'bootc': '.bo'
}


FILE_GENERATED_HASH_COMMENT = '# THIS FILE IS GENERATED'

REMEDIATION_CONFIG_KEYS = ['complexity', 'disruption', 'platform', 'reboot',
                           'strategy']
REMEDIATION_ELM_KEYS = ['complexity', 'disruption', 'reboot', 'strategy']

RemediationObject = namedtuple('remediation', ['contents', 'config'])


def is_supported_filename(remediation_type, filename):
    """
    Checks if filename has a supported extension for remediation_type.

    Exits when remediation_type is of an unknown type.
    """
    if remediation_type in REMEDIATION_TO_EXT_MAP:
        return filename.endswith(REMEDIATION_TO_EXT_MAP[remediation_type])

    sys.stderr.write("ERROR: Unknown remediation type '%s'!\n"
                     % (remediation_type))
    sys.exit(1)


def split_remediation_content_and_metadata(fix_file):
    remediation_contents = []
    config = defaultdict(lambda: None)

    # Assignment automatically escapes shell characters for XML
    for line in fix_file.splitlines():
        if line.startswith(FILE_GENERATED_HASH_COMMENT):
            continue

        if line.startswith('#') and line.count('=') == 1:
            (key, value) = line.strip('#').split('=')
            if key.strip() in REMEDIATION_CONFIG_KEYS:
                config[key.strip()] = value.strip()
                continue

        # If our parsed line wasn't a config item, add it to the
        # returned file contents. This includes when the line
        # begins with a '#' and contains an equals sign, but
        # the "key" isn't one of the known keys from
        # REMEDIATION_CONFIG_KEYS.
        remediation_contents.append(line)

    contents = "\n".join(remediation_contents)
    return RemediationObject(contents=contents, config=config)


def parse_from_file_with_jinja(file_path, env_yaml):
    """
    Parses a remediation from a file. As remediations contain jinja macros,
    we need a env_yaml context to process these. In practice, no remediations
    use jinja in the configuration, so for extracting only the configuration,
    env_yaml can be an abritrary product.yml dictionary.

    If the logic of configuration parsing changes significantly, please also
    update ssg.fixes.parse_platform(...).
    """

    fix_file = jinja_process_file(file_path, env_yaml)
    return split_remediation_content_and_metadata(fix_file)


def parse_from_file_without_jinja(file_path):
    """
    Parses a remediation from a file. Doesn't process the Jinja macros.
    This function is useful in build phases in which all the Jinja macros
    are already resolved.
    """
    with open(file_path, "r") as f:
        f_str = f.read()
        return split_remediation_content_and_metadata(f_str)


class Remediation(object):
    def __init__(self, file_path, remediation_type):
        self.file_path = file_path
        self.local_env_yaml = dict()

        self.metadata = defaultdict(lambda: None)

        self.remediation_type = remediation_type
        self.associated_rule = None

    def associate_rule(self, rule_obj):
        self.associated_rule = rule_obj
        self.expand_env_yaml_from_rule()

    def expand_env_yaml_from_rule(self):
        if not self.associated_rule:
            return

        self.local_env_yaml["rule_title"] = self.associated_rule.title
        self.local_env_yaml["rule_id"] = self.associated_rule.id_
        self.local_env_yaml["cce_identifiers"] = self.associated_rule.identifiers

    def parse_from_file_with_jinja(self, env_yaml, cpe_platforms):
        return parse_from_file_with_jinja(self.file_path, env_yaml)

    def get_inherited_cpe_platform_names(self):
        inherited_cpe_platform_names = set()
        if self.associated_rule:
            # There can be repeated inherited platforms and rule platforms
            inherited_cpe_platform_names.update(self.associated_rule.inherited_cpe_platform_names)
        return inherited_cpe_platform_names

    def get_rule_specific_cpe_platform_names(self):
        rule_specific_cpe_platform_names = set()
        inherited_cpe_platform_names = self.get_inherited_cpe_platform_names()
        if self.associated_rule and self.associated_rule.cpe_platform_names is not None:
            rule_specific_cpe_platform_names = {
                p for p in self.associated_rule.cpe_platform_names
                if p not in inherited_cpe_platform_names}
        return rule_specific_cpe_platform_names

    def _get_stripped_conditional(self, language, platform):
        conditional = platform.get_remediation_conditional(language)
        if conditional is not None:
            stripped_conditional = conditional.strip()
            if stripped_conditional:
                return stripped_conditional
        return None

    def get_stripped_conditionals(self, language, cpe_platform_names, cpe_platforms):
        """
        collect conditionals of platforms defined by cpe_platform_names
        and strip them of white spaces
        """
        stripped_conditionals = []
        for p in cpe_platform_names:
            platform = cpe_platforms[p]
            maybe_stripped_conditional = self._get_stripped_conditional(language, platform)
            if maybe_stripped_conditional is not None:
                stripped_conditionals.append(maybe_stripped_conditional)
        return stripped_conditionals

    def get_rule_specific_conditionals(self, language, cpe_platforms):
        cpe_platform_names = self.get_rule_specific_cpe_platform_names()
        return self.get_stripped_conditionals(language, cpe_platform_names, cpe_platforms)

    def get_inherited_conditionals(self, language, cpe_platforms):
        cpe_platform_names = self.get_inherited_cpe_platform_names()
        return self.get_stripped_conditionals(language, cpe_platform_names, cpe_platforms)


def process(remediation, env_yaml, cpe_platforms):
    """
    Process a fix, and return the processed fix iff the file is of a valid
    extension for the remediation type and the fix is valid for the current
    product.

    Note that platform is a required field in the contents of the fix.
    """
    if not is_supported_filename(remediation.remediation_type, remediation.file_path):
        return

    result = remediation.parse_from_file_with_jinja(env_yaml, cpe_platforms)
    platforms = result.config['platform']

    if not platforms:
        raise RuntimeError(
            "The '%s' remediation script does not contain the "
            "platform identifier!" % (remediation.file_path))

    for platform in platforms.split(","):
        if platform.strip() != platform:
            msg = (
                "Comma-separated '{platform}' platforms "
                "in '{remediation_file}' contains whitespace."
                .format(platform=platforms, remediation_file=remediation.file_path))
            raise ValueError(msg)

    product = env_yaml["product"]
    if utils.is_applicable_for_product(platforms, product):
        return result

    return None


class BashRemediation(Remediation):
    def __init__(self, file_path):
        super(BashRemediation, self).__init__(file_path, "bash")

    def parse_from_file_with_jinja(self, env_yaml, cpe_platforms):
        self.local_env_yaml.update(env_yaml)
        result = super(BashRemediation, self).parse_from_file_with_jinja(
            self.local_env_yaml, cpe_platforms)

        # Avoid platform wrapping empty fix text
        # Remediations can be empty when a Jinja macro or conditional
        # renders no fix text for a product
        stripped_fix_text = result.contents.strip()
        if stripped_fix_text == "":
            return result

        inherited_conditionals = sorted(super(
            BashRemediation, self).get_inherited_conditionals("bash", cpe_platforms))
        rule_specific_conditionals = sorted(super(
            BashRemediation, self).get_rule_specific_conditionals("bash", cpe_platforms))
        if inherited_conditionals or rule_specific_conditionals:
            wrapped_fix_text = ["# Remediation is applicable only in certain platforms"]

            all_conditions = ""
            if inherited_conditionals:
                all_conditions += " && ".join(inherited_conditionals)
            if rule_specific_conditionals:
                if all_conditions:
                    all_conditions += " && { " + " || ".join(rule_specific_conditionals) + "; }"
                else:
                    all_conditions = " || ".join(rule_specific_conditionals)
            wrapped_fix_text.append("if {0}; then".format(all_conditions))
            wrapped_fix_text.append("")
            # It is possible to indent the original body of the remediation with textwrap.indent(),
            # however, it is not supported by python2, and there is a risk of breaking remediations
            # For example, remediations with a here-doc block could be affected.
            wrapped_fix_text.append("{0}".format(stripped_fix_text))
            wrapped_fix_text.append("")
            wrapped_fix_text.append("else")
            wrapped_fix_text.append(
                "    >&2 echo 'Remediation is not applicable, nothing was done'")
            wrapped_fix_text.append("fi")

            result = RemediationObject(contents="\n".join(wrapped_fix_text), config=result.config)

        return result


class AnsibleRemediation(Remediation):
    def __init__(self, file_path):
        super(AnsibleRemediation, self).__init__(
            file_path, "ansible")

        self.body = None

    def parse_from_file_with_jinja(self, env_yaml, cpe_platforms):
        self.local_env_yaml.update(env_yaml)
        result = super(AnsibleRemediation, self).parse_from_file_with_jinja(
            self.local_env_yaml, cpe_platforms)

        if not self.associated_rule:
            return result

        parsed = ssg.yaml.ordered_load(result.contents)

        self.update(parsed, result.config, cpe_platforms)

        updated_yaml_text = ssg.yaml.ordered_dump(
            parsed, None, default_flow_style=False)
        result = result._replace(contents=updated_yaml_text)

        self.body = parsed
        self.metadata = result.config

        return result

    def update_tags_from_config(self, to_update, config):
        tags = to_update.get("tags", [])
        if "strategy" in config:
            tags.append("{0}_strategy".format(config["strategy"]))
        if "complexity" in config:
            tags.append("{0}_complexity".format(config["complexity"]))
        if "disruption" in config:
            tags.append("{0}_disruption".format(config["disruption"]))
        if "reboot" in config:
            if config["reboot"] == "true":
                reboot_tag = "reboot_required"
            else:
                reboot_tag = "no_reboot_needed"
            tags.append(reboot_tag)
        to_update["tags"] = sorted(tags)

    def update_tags_from_rule(self, to_update):
        if not self.associated_rule:
            raise RuntimeError("The Ansible snippet has no rule loaded.")

        tags = to_update.get("tags", [])
        tags.insert(0, "{0}_severity".format(self.associated_rule.severity))
        tags.insert(0, self.associated_rule.id_)

        cce_num = self._get_cce()
        if cce_num:
            tags.append("{0}".format(cce_num))

        refs = self.get_references()
        tags.extend(refs)
        to_update["tags"] = sorted(tags)

    def _get_cce(self):
        return self.associated_rule.identifiers.get("cce", None)

    def get_references(self):
        if not self.associated_rule:
            raise RuntimeError("The Ansible snippet has no rule loaded.")

        result = []
        for ref_class, prefix in constants.REF_PREFIX_MAP.items():
            refs = self._get_rule_reference(ref_class)
            result.extend(["{prefix}-{value}".format(prefix=prefix, value=v) for v in refs])
        return result

    def _get_rule_reference(self, ref_class):
        refs = self.associated_rule.references.get(ref_class, [])
        return refs

    def inject_package_facts_task(self, parsed_snippet):
        """ Injects a package_facts task only if
            the snippet has a task with a when clause with ansible_facts.packages,
            and the snippet doesn't already have a package_facts task
        """
        has_package_facts_task = False
        has_ansible_facts_packages_clause = False

        for p_task in parsed_snippet:
            # We are only interested in the OrderedDicts, which represent Ansible tasks
            if not isinstance(p_task, dict):
                continue

            if "package_facts" in p_task:
                has_package_facts_task = True

            # When clause of the task can be string or a list, lets normalize to list
            task_when = p_task.get("when", "")
            if type(task_when) is str:
                task_when = [task_when]
            for when in task_when:
                if "ansible_facts.packages" in when:
                    has_ansible_facts_packages_clause = True

        if has_ansible_facts_packages_clause and not has_package_facts_task:
            facts_task = OrderedDict([
                ('name', 'Gather the package facts'),
                ('package_facts', {'manager': 'auto'})
            ])
            parsed_snippet.insert(0, facts_task)

    def update_when_from_rule(self, to_update, cpe_platforms):
        additional_when = []
        inherited_conditionals = sorted(super(
            AnsibleRemediation, self).get_inherited_conditionals("ansible", cpe_platforms))
        rule_specific_conditionals = sorted(super(
            AnsibleRemediation, self).get_rule_specific_conditionals("ansible", cpe_platforms))
        # Remove conditionals related to package CPEs if the updated task collects package facts
        if "package_facts" in to_update:
            inherited_conditionals = filter(
                lambda c: "in ansible_facts.packages" not in c,
                inherited_conditionals)
            rule_specific_conditionals = filter(
                lambda c: "in ansible_facts.packages" not in c, rule_specific_conditionals)

        if inherited_conditionals:
            additional_when.extend(inherited_conditionals)

        if rule_specific_conditionals:
            additional_when.append(" or ".join(rule_specific_conditionals))

        to_update.setdefault("when", "")
        new_when = ssg.yaml.update_yaml_list_or_string(to_update["when"], additional_when,
                                                       prepend=True)
        if not new_when:
            to_update.pop("when")
        else:
            to_update["when"] = new_when

    def update(self, parsed, config, cpe_platforms):
        # We split the remediation update in three steps

        # 1. Update the when clause
        for p in parsed:
            if not isinstance(p, dict):
                continue
            self.update_when_from_rule(p, cpe_platforms)

        # 2. Inject any extra task necessary
        self.inject_package_facts_task(parsed)

        # 3. Add tags to all tasks, including the ones we have injected
        for p in parsed:
            if not isinstance(p, dict):
                continue
            self.update_tags_from_config(p, config)
            self.update_tags_from_rule(p)

    @classmethod
    def from_snippet_and_rule(cls, snippet_fname, rule_fname):
        if os.path.isfile(snippet_fname) and os.path.isfile(rule_fname):
            result = cls(snippet_fname)
            try:
                rule_obj = ssg.build_yaml.Rule.from_yaml(rule_fname)
                result.associate_rule(rule_obj)
            except ssg.yaml.DocumentationNotComplete:
                # Happens on non-debug build when a rule is "documentation-incomplete"
                return None
            return result


class AnacondaRemediation(Remediation):
    def __init__(self, file_path):
        super(AnacondaRemediation, self).__init__(
            file_path, "anaconda")


class PuppetRemediation(Remediation):
    def __init__(self, file_path):
        super(PuppetRemediation, self).__init__(
            file_path, "puppet")


class IgnitionRemediation(Remediation):
    def __init__(self, file_path):
        super(IgnitionRemediation, self).__init__(
            file_path, "ignition")


class KubernetesRemediation(Remediation):
    def __init__(self, file_path):
        super(KubernetesRemediation, self).__init__(
              file_path, "kubernetes")


class BlueprintRemediation(Remediation):
    """
    This provides class for OSBuild Blueprint remediations
    """
    def __init__(self, file_path):
        super(BlueprintRemediation, self).__init__(
            file_path, "blueprint")


class KickstartRemediation(Remediation):
    """
    This provides class for Kickstart remediations
    """
    def __init__(self, file_path):
        super(KickstartRemediation, self).__init__(
            file_path, "kickstart")


class BootcRemediation(Remediation):
    """
    This provides class for Bootc remediations
    """
    def __init__(self, file_path):
        super(BootcRemediation, self).__init__(
            file_path, "bootc")


REMEDIATION_TO_CLASS = {
    'anaconda': AnacondaRemediation,
    'ansible': AnsibleRemediation,
    'bash': BashRemediation,
    'puppet': PuppetRemediation,
    'ignition': IgnitionRemediation,
    'kubernetes': KubernetesRemediation,
    'blueprint': BlueprintRemediation,
    'kickstart': KickstartRemediation,
    'bootc': BootcRemediation,
}


def write_fix_to_file(fix, file_path):
    """
    Writes a single fix to the given file path.
    """
    fix_contents, config = fix
    with open(file_path, "w") as f:
        for k, v in config.items():
            f.write("# %s = %s\n" % (k, v))
        f.write(fix_contents)


def get_rule_dir_remediations(dir_path, remediation_type, product=None):
    """
    Gets a list of remediations of type remediation_type contained in a
    rule directory. If product is None, returns all such remediations.
    If product is not None, returns applicable remediations in order of
    priority:

        {{{ product }}}.ext -> shared.ext

    Only returns remediations which exist.
    """

    if not rules.is_rule_dir(dir_path):
        return []

    remediations_dir = os.path.join(dir_path, remediation_type)
    has_remediations_dir = os.path.isdir(remediations_dir)
    ext = REMEDIATION_TO_EXT_MAP[remediation_type]
    if not has_remediations_dir:
        return []

    # Two categories of results: those for a product and those that are
    # shared to multiple products. Within common results, there's two types:
    # those shared to multiple versions of the same type (added up front) and
    # those shared across multiple product types (e.g., RHEL and Ubuntu).
    product_results = []
    common_results = []
    for remediation_file in sorted(os.listdir(remediations_dir)):
        file_name, file_ext = os.path.splitext(remediation_file)
        remediation_path = os.path.join(remediations_dir, remediation_file)

        if file_ext == ext and rules.applies_to_product(file_name, product):
            # rules.applies_to_product ensures we only have three entries:
            # 1. shared
            # 2. <product>
            # 3. <product><version>
            #
            # Note that the product variable holds <product><version>.
            if file_name == 'shared':
                # Shared are the lowest priority items, add them to the end
                # of the common results.
                common_results.append(remediation_path)
            elif file_name != product:
                # Here, the filename is a subset of the product, but isn't
                # the full product. Product here is both the product name
                # (e.g., ubuntu) and its version (2004). Filename could be
                # either "ubuntu" or "ubuntu2004" so we want this branch
                # to trigger when it is the former, not the latter. It is
                # the highest priority of common results, so insert it
                # before any shared ones.
                common_results.insert(0, remediation_path)
            else:
                # Finally, this must be product-specific result.
                product_results.append(remediation_path)

    # Combine the two sets in priority order.
    return product_results + common_results


def expand_xccdf_subs(fix, remediation_type):
    """Expand the respective populate keywords of each
    remediation type with an <xccdf:sub> element

    This routine translates any instance of the '`type`-populate' keyword in
    the form of:

            (`type`-populate variable_name)

    where `type` can be either ansible, puppet, anaconda or bash, into

            <sub idref="variable_name"/>

    """

    if fix is not None:
        fix_text = fix.text
    else:
        return
    if remediation_type == "ignition":
        return
    elif remediation_type == "kubernetes":
        return
    elif remediation_type == "blueprint":
        pattern = r'\(blueprint-populate\s*(\S+)\)'
    elif remediation_type == "ansible":

        if "(ansible-populate " in fix_text:
            raise RuntimeError(
                "(ansible-populate VAR) has been deprecated. Please use "
                "(xccdf-var VAR) instead. Keep in mind that the latter will "
                "make an ansible variable out of XCCDF Value as opposed to "
                "substituting directly."
            )

        # If you change this string make sure it still matches the pattern
        # defined in OpenSCAP. Otherwise you break variable handling in
        # 'oscap xccdf generate fix' and the variables won't be customizable!
        # https://github.com/OpenSCAP/openscap/blob/1.2.17/src/XCCDF_POLICY/xccdf_policy_remediate.c#L588
        #   const char *pattern =
        #     "- name: XCCDF Value [^ ]+ # promote to variable\n  set_fact:\n"
        #     "    ([^:]+): (.+)\n  tags:\n    - always\n";
        # We use !!str typecast to prevent treating values as different types
        # eg. yes as a bool or 077 as an octal number
        fix_text = re.sub(
            r"- \(xccdf-var\s+(\S+)\)",
            r"- name: XCCDF Value \1 # promote to variable\n"
            r"  set_fact:\n"
            r"    \1: !!str (ansible-populate \1)\n"
            r"  tags:\n"
            r"    - always",
            fix_text
        )

        pattern = r'\(ansible-populate\s*(\S+)\)'

    elif remediation_type == "puppet":
        pattern = r'\(puppet-populate\s*(\S+)\)'

    elif remediation_type == "anaconda":
        pattern = r'\(anaconda-populate\s*(\S+)\)'

    elif remediation_type == "bash":
        pattern = r'\(bash-populate\s*(\S+)\)'
    elif remediation_type == "kickstart":
        pattern = r'\(kickstart-populate\s*(\S+)\)'
    elif remediation_type == "bootc":
        pattern = r'\(bootc-populate\s*(\S+)\)'
    else:
        sys.stderr.write("Unknown remediation type '%s'\n" % (remediation_type))
        sys.exit(1)

    # we will get list what looks like
    # [text, varname, text, varname, ..., text]
    parts = re.split(pattern, fix_text)

    fix.text = parts[0]  # add first "text"
    for index in range(1, len(parts), 2):
        varname = parts[index]
        text_between_vars = parts[index + 1]

        # we cannot combine elements and text easily
        # so text is in ".tail" of element
        xccdfvarsub = ElementTree.SubElement(
            fix, "{%s}sub" % XCCDF12_NS, idref=constants.OSCAP_VALUE + varname)
        xccdfvarsub.tail = text_between_vars
        xccdfvarsub.set("use", "legacy")


def load_compiled_remediations(fixes_dir):
    if not os.path.isdir(fixes_dir):
        raise RuntimeError(
            "Directory with compiled fixes '%s' does not exist" % fixes_dir)
    all_remediations = defaultdict(dict)
    for language in os.listdir(fixes_dir):
        language_dir = os.path.join(fixes_dir, language)
        if not os.path.isdir(language_dir):
            raise RuntimeError(
                "Can't find the '%s' directory with fixes for %s" %
                (language_dir, language))
        for filename in sorted(os.listdir(language_dir)):
            file_path = os.path.join(language_dir, filename)
            rule_id, _ = os.path.splitext(filename)
            remediation = parse_from_file_without_jinja(file_path)
            all_remediations[rule_id][language] = remediation
    return all_remediations



#!/usr/bin/python3

from __future__ import print_function

import argparse
import json
import os
import re
from pathlib import Path
import sys
import xml.etree.ElementTree as ET
import yaml

import ssg.build_yaml
import ssg.controls
import ssg.environment
import ssg.rules
from ssg.utils import mkdir_p
import ssg.yaml


SSG_ROOT = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))
BUILD_OUTPUT = os.path.join(SSG_ROOT, "build", "stig_control.yml")
RULES_JSON = os.path.join(SSG_ROOT, "build", "rule_dirs.json")
BUILD_CONFIG = os.path.join(SSG_ROOT, "build", "build_config.yml")


def check_output(output: str) -> None:
    pat = re.compile(r'.*\/?[a-z_0-9]+\.yml')
    if not pat.match(output):
        sys.stderr.write('Output must only contain lowercase letters, underscores, and numbers.'
                         ' The file must also end with .yml\n')
        exit(1)


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser()
    parser.add_argument("-r", "--root", type=str, action="store", default=SSG_ROOT,
                        help="Path to SSG root directory (defaults to %s)" % SSG_ROOT)
    parser.add_argument("-o", "--output", type=str, action="store", default=BUILD_OUTPUT,
                        help=f"File to write yaml output to (defaults to {BUILD_OUTPUT}). "
                             f"Must end in '.yml' and only contain "
                             f"lowercase letters, underscores, and numbers.")
    parser.add_argument("-p", "--product", type=str, action="store", required=True,
                        help="What product to get STIGs for")
    parser.add_argument("-m", "--manual", type=str, action="store", required=True,
                        help="Path to XML XCCDF manual file to use as the source of the STIGs")
    parser.add_argument("-g", "--srg-control", type=str, action="store", default=None,
                        help="Path to the SRG control file relevant to the STIG")
    parser.add_argument("-j", "--json", type=str, action="store", default=RULES_JSON,
                        help=f"Path to the rules_dir.json (defaults to {RULES_JSON})")
    parser.add_argument("-c", "--build-config-yaml", default=BUILD_CONFIG,
                        help="YAML file with information about the build configuration")
    parser.add_argument("-ref", "--reference", type=str, default="stigid",
                        help="Reference system to check for, defaults to stigid")
    parser.add_argument('-s', '--split', action='store_true',
                        help='Splits the each ID into its own file.')

    args = parser.parse_args()
    check_output(args.output)
    return args


def handle_rule_yaml(args, rule_id, rule_dir, guide_dir, env_yaml):
    rule_obj = {'id': rule_id, 'dir': rule_dir, 'guide': guide_dir}
    rule_file = ssg.rules.get_rule_dir_yaml(rule_dir)

    rule_yaml = ssg.build_yaml.Rule.from_yaml(rule_file, env_yaml=env_yaml)
    rule_yaml.normalize(args.product)
    rule_obj['references'] = rule_yaml.references
    return rule_obj


def get_platform_rules(args):
    rules_json_file = open(args.json, 'r')
    rules_json = json.load(rules_json_file)
    platform_rules = list()
    for rule in rules_json.values():
        if args.product in rule['products']:
            platform_rules.append(rule)
    if not rules_json_file.closed:
        rules_json_file.close()
    return platform_rules


def map_rule_id_to_ref(rule_id, refs, known_rules):
    for ref in refs:
        if ref in known_rules:
            known_rules[ref].append(rule_id)
        else:
            known_rules[ref] = [rule_id]


def get_implemented_stigs(args, env_yaml):
    platform_rules = get_platform_rules(args)

    known_rules = dict()
    for rule in platform_rules:
        try:
            rule_obj = handle_rule_yaml(args, rule['id'],
                                        rule['dir'], rule['guide'], env_yaml)
        except ssg.yaml.DocumentationNotComplete:
            sys.stderr.write('Rule %s throw DocumentationNotComplete' % rule['id'])
            # Happens on non-debug build when a rule is "documentation-incomplete"
            continue

        refs = rule_obj['references'].get(args.reference, [])
        map_rule_id_to_ref(rule['id'], refs, known_rules)
    return known_rules


def check_files(args):
    if not os.path.exists(args.json):
        sys.stderr.write('Unable to find %s\n' % args.json)
        sys.stderr.write('Hint: run ./utils/rule_dir_json.py\n')
        exit(-1)

    if not os.path.exists(args.build_config_yaml):
        sys.stderr.write('Unable to find %s\n' % args.build_config_yaml)
        sys.stderr.write('Hint: build the project,\n')
        exit(-1)


def get_extra_srgs(rule, ns) -> list:
    pattern = re.compile(r'SRG-[A-Z]{2,}-\d{5,}-[A-Z]{3,}-\d{5,}')

    description = rule.find('checklist:description', ns).text
    return pattern.findall(description)


def get_rules_for_control(stig_id, known_rules, srgs, srg_controls):
    # Add any known rule with the same STIG ID reference
    rule_set = set()
    if stig_id in known_rules.keys():
        rule_set.update(known_rules.get(stig_id))

    # Let's also add any rule selected in the SRG control file
    if srg_controls:
        for srg in srgs:
            rule_set.update(srg_controls.get_control(srg).rules)

    return sorted(list(rule_set))


def get_controls(known_rules, ns, root, srg_controls=None) -> list:
    controls = list()
    for group in root.findall('checklist:Group', ns):
        # There is always at least one SRG associated
        srgs = [group.find('checklist:title', ns).text]

        for stig in group.findall('checklist:Rule', ns):
            stig_id = stig.find('checklist:version', ns).text
            # Add any other SRG associated mentioned in description
            srgs += get_extra_srgs(stig, ns)
            control = dict()
            control['id'] = stig_id
            control['levels'] = [stig.attrib['severity']]
            control['title'] = stig.find('checklist:title', ns).text
            control['rules'] = get_rules_for_control(stig_id, known_rules, srgs, srg_controls)
            if len(control['rules']) > 0:
                control['status'] = 'automated'
            else:
                control['status'] = 'pending'

            controls.append(control)
    return controls


def get_disa_stig_version(root, ns):
    version = root.find('checklist:version', ns).text
    release_string = root.find('checklist:plain-text[@id="release-info"]', ns).text
    release = re.match(r'Release: (\d+) Benchmark', release_string)
    return f"V{version}R{release.group(1)}"


def main():
    args = parse_args()
    check_files(args)

    product_dir = os.path.join(args.root, "products", args.product)
    product_yaml_path = os.path.join(product_dir, "product.yml")
    env_yaml = ssg.environment.open_environment(
        args.build_config_yaml, product_yaml_path, os.path.join(args.root, "product_properties"))

    ns = {'checklist': 'http://checklists.nist.gov/xccdf/1.1'}
    known_rules = get_implemented_stigs(args, env_yaml)
    tree = ET.parse(args.manual)
    root = tree.getroot()
    output = dict()
    output['policy'] = root.find('checklist:title', ns).text
    output['title'] = root.find('checklist:title', ns).text
    output['id'] = 'stig_%s' % args.product
    output['version'] = get_disa_stig_version(root, ns)
    output['source'] = 'https://public.cyber.mil/stigs/downloads/'
    output['reference_type'] = "stigid"
    output['product'] = args.product
    output['levels'] = list()
    for level in ['high', 'medium', 'low']:
        output['levels'].append({'id': level})

    srg_controls = None
    if args.srg_control:
        srg_controls = ssg.controls.Policy(args.srg_control, env_yaml)
        srg_controls.load()
    controls = get_controls(known_rules, ns, root, srg_controls)

    if args.split:
        with open(args.output, 'w') as f:
            f.write(yaml.dump(output, sort_keys=False))
        print(f'Wrote main control file to {args.output}')
        output_path = Path(args.output)
        output_dir_name = output_path.stem
        output_root = output_path.parent
        output_dir = os.path.join(output_root, output_dir_name)
        mkdir_p(output_dir)
        for control in controls:
            out = dict()
            out['controls'] = [control, ]
            filename = f"{control['id']}.yml"
            output_filename = os.path.join(output_dir, filename)
            with open(output_filename, 'w') as f:
                f.write(yaml.dump(out, sort_keys=False))
        print(f'Wrote SRG files to {output_dir}')
        exit(0)
    else:
        output['controls'] = controls
        with open(args.output, 'w') as f:
            f.write(yaml.dump(output, sort_keys=False))
        print(f'Wrote all SRGs out to {args.output}')


if __name__ == "__main__":
    main()



"""
The rule_yaml module provides various utility functions for handling YAML files
containing Jinja macros, without having to parse the macros.
"""

from __future__ import absolute_import
from __future__ import print_function

import os
import sys
from collections import namedtuple, defaultdict
import yaml

from .rules import get_rule_dir_yaml
from .utils import read_file_list


def find_section_lines(file_contents, sec):
    """
    Parses the given file_contents as YAML to find the section with the given identifier.

    Note that this does not call into the yaml library and thus correctly handles Jinja macros at
    the expense of not being a strictly valid yaml parsing.

    Args:
        file_contents (list of str): The contents of the file, split into lines.
        sec (str): The identifier of the section to find.

    Returns:
        list of namedtuple: A list of namedtuples (start, end) representing the lines where the
                            section exists.
    """
    # Hack to find a global key ("section"/sec) in a YAML-like file.
    # All indented lines until the next global key are included in the range.
    # For example:
    #
    # 0: not_it:
    # 1:     - value
    # 2: this_one:
    # 3:      - 2
    # 4:      - 5
    # 5:
    # 6: nor_this:
    #
    # for the section "this_one", the result [(2, 5)] will be returned.
    # Note that multiple sections may exist in a file and each will be
    # identified and returned.
    section = namedtuple('section', ['start', 'end'])

    sec_ranges = []
    sec_id = sec + ":"
    sec_len = len(sec_id)
    end_num = len(file_contents)
    line_num = 0

    while line_num < end_num:
        if len(file_contents[line_num]) >= sec_len:
            if file_contents[line_num][0:sec_len] == sec_id:
                begin = line_num
                line_num += 1
                while line_num < end_num:
                    nonempty_line = file_contents[line_num]
                    if nonempty_line and file_contents[line_num][0] != ' ':
                        break
                    line_num += 1

                end = line_num - 1
                sec_ranges.append(section(begin, end))
        line_num += 1

    return sec_ranges


def add_key_value(contents, key, start_line, new_value):
    """
    Adds a new key-value pair to the contents at a specified line.

    Does not modify the value of contents.

    Args:
        contents (list of str): The original list of strings representing the contents.
        key (str): The key to be added.
        start_line (int): The line number after which the new key-value pair should be added.
        new_value (str): The value associated with the key.

    Returns:
        list of str: A new list of strings with the key-value pair added and a blank line afterwards.
    """
    new_contents = contents[:start_line]
    new_contents.append("%s: %s" % (key, new_value))
    new_contents.append("")
    new_contents.extend(contents[start_line:])

    return new_contents


def update_key_value(contents, key, old_value, new_value):
    """
    Find a key in the contents of a file and replace its value with a new value, returning the
    resulting file contents.

    This function validates that the old value is constant and hasn't changed since parsing its
    value. Does not modify the original contents.

    Args:
        contents (list of str): The contents of the file as a list of strings.
        key (str): The key whose value needs to be updated.
        old_value (str): The current value associated with the key.
        new_value (str): The new value to replace the old value.

    Returns:
        list of str: The updated contents of the file.

    Raises:
        ValueError: If the key cannot be found in the given contents or if the old value does not
                    match the current value associated with the key.
    """
    new_contents = contents[:]
    old_line = key + ": " + old_value
    updated = False

    for line_num in range(0, len(new_contents)):
        line = new_contents[line_num]
        if line == old_line:
            new_contents[line_num] = key + ": " + new_value
            updated = True
            break

    if not updated:
        raise ValueError("For key:%s, cannot find the old value (%s) in the given "
                         "contents." % (key, old_value))

    return new_contents


def remove_lines(contents, lines):
    """
    Remove the specified lines from the contents.

    This function takes the contents of a file and a range of lines to be removed, and returns the
    new contents with those lines removed. The original contents are not modified.

    Args:
        contents (list of str): The contents of the file as a list of lines.
        lines (slice): A slice object representing the range of lines to be removed.

    Returns:
        list of str: The new contents with the specified lines removed.
    """
    new_contents = contents[:lines.start]
    new_contents.extend(contents[lines.end+1:])
    return new_contents


def parse_from_yaml(file_contents, lines):
    """
    Parse the given line range as a YAML, returning the parsed object.

    Args:
        file_contents (list of str): The contents of the file as a list of strings.
        lines (slice): A slice object indicating the start and end lines to parse.

    Returns:
        object: The parsed YAML object.
    """
    new_file_arr = file_contents[lines.start:lines.end + 1]
    new_file = "\n".join(new_file_arr)
    return yaml.load(new_file, Loader=yaml.Loader)


def get_yaml_contents(rule_obj):
    """
    From a rule_obj description, return a namedtuple of (path, contents).

    Args:
        rule_obj (dict): A dictionary containing information about the rule.
                         It must have the keys 'dir' and 'id'.

    Returns:
        namedtuple: A namedtuple with 'path' as the path to the rule YAML file and 'contents' as
                    the list of lines in the file.

    Raises:
        ValueError: If the YAML file does not exist for the given rule_id.
    """
    file_description = namedtuple('file_description', ('path', 'contents'))

    yaml_file = get_rule_dir_yaml(rule_obj['dir'])
    if not os.path.exists(yaml_file):
        raise ValueError("Error: yaml file does not exist for rule_id:%s" %
                         rule_obj['id'], file=sys.stderr)

    yaml_contents = read_file_list(yaml_file)

    return file_description(yaml_file, yaml_contents)


def get_section_lines(file_path, file_contents, key_name):
    """
    From the given file_path and file_contents, find the lines describing the section key_name and
    returns the line range of the section.

    Args:
        file_path (str): The path to the file being analyzed.
        file_contents (str): The contents of the file as a string.
        key_name (str): The name of the section to find within the file contents.

    Returns:
        tuple: A tuple representing the start and end line numbers of the section if found.
        None: If the section is not found.

    Raises:
        ValueError: If multiple instances of the section are found in the file.
    """
    section = find_section_lines(file_contents, key_name)

    if len(section) > 1:
        raise ValueError("Multiple instances (%d) of %s in %s; refusing to modify file." %
                         (len(section), key_name, file_path), file=sys.stderr)

    elif len(section) == 1:
        return section[0]

    return None


def has_duplicated_subkeys(file_path, file_contents, sections):
    """
    Checks whether a section has duplicated keys in a YAML file.

    Note that these duplicated keys are silently ignored by the YAML parser used.

    Args:
        file_path (str): The path to the YAML file.
        file_contents (list of str): The contents of the YAML file as a list of lines.
        sections (str or list of str): The section or list of sections to check for duplicated keys.

    Returns:
        bool: True if any section has duplicated keys, False otherwise.
    """
    if isinstance(sections, str):
        sections = [sections]

    for section in sections:
        # Get the lines in the file which match this section. If none exists,
        # it should be safe to silently ignore it. Clearly if the section
        # exists, there are no duplicated sections.
        section_range = get_section_lines(file_path, file_contents, section)
        if not section_range:
            continue

        # Get the YAML parser's version of events. :-)
        parsed_section = parse_from_yaml(file_contents, section_range)

        # Sort the YAML parser's subkeys.
        parent_key = list(parsed_section.keys())[0]
        if not parsed_section[parent_key]:
            continue
        subkeys = parsed_section[parent_key].keys()

        # Create a dictionary for counting them.
        subkey_counts = defaultdict(lambda: 0)

        # Iterate over the lines, see if they match a known key. Ignore the
        # first line (as it is the section header).
        for line_num in range(section_range.start+1, section_range.end):
            line = file_contents[line_num]
            if not line:
                continue

            # We'll be lazy for the time being. Iterate over all keys.
            for key in subkeys:
                our_key = ' ' + key + ':'
                if our_key in line:
                    subkey_counts[our_key] += 1
                    if subkey_counts[our_key] > 1:
                        print("Duplicated key " + our_key + " in " + section + " of " + file_path)
                        return True

    return False


def sort_section_keys(file_path, file_contents, sections, sort_func=None):
    """
    Sort subkeys in a YAML file's section.

    Args:
        file_path (str): The path to the YAML file.
        file_contents (list of str): The contents of the YAML file as a list of lines.
        sections (str or list of str): The section or sections whose subkeys need to be sorted.
        sort_func (callable, optional): A function to determine the sort order of the subkeys.
                                        If None, the subkeys are sorted in ascending order.

    Returns:
        list of str: The modified contents of the YAML file with sorted subkeys in the specified
                     sections.

    Raises:
        ValueError: If a duplicated key is found within the same section.
        AssertionError: If the section contains more than one parent key or if a subkey line
                        cannot be found.
    """
    if isinstance(sections, str):
        sections = [sections]

    new_contents = file_contents[:]

    for section in sections:
        section_range = get_section_lines(file_path, new_contents, section)
        if not section_range:
            continue

        # Start by parsing the lines as YAML.
        parsed_section = parse_from_yaml(new_contents, section_range)

        # Ignore the section header. This header is included in the start range,
        # so just increment by one.
        start_offset = 1
        while not new_contents[section_range.start + start_offset].strip():
            start_offset += 1

        # Ignore any trailing empty lines.
        end_offset = 0
        while not new_contents[section_range.end - end_offset].strip():
            end_offset += 1

        # Validate we only have a single section.
        assert len(parsed_section.keys()) == 1

        # Sort the parsed subkeys.
        parent_key = list(parsed_section.keys())[0]
        if not parsed_section[parent_key]:
            continue
        subkeys = sorted(parsed_section[parent_key].keys(), key=sort_func)

        # Don't bother if there are zero or one subkeys. Sorting order thus
        # doesn't matter.
        if not subkeys or len(subkeys) == 1:
            continue

        # Now we need to map sorted subkeys onto lines in the new contents,
        # so we can re-order them appropriately. We'll assume the section is
        # small so we'll do it in O(n^2).
        subkey_mapping = dict()
        for key in subkeys:
            our_line = None
            spaced_key = ' ' + key + ':'
            tabbed_key = '\t' + key + ':'
            range_start = section_range.start + start_offset
            range_end = section_range.end - end_offset + 1
            for line_num in range(range_start, range_end):
                this_line = new_contents[line_num]
                if spaced_key in this_line or tabbed_key in this_line:
                    if our_line:
                        # Not supposed to be possible to have multiple keys
                        # matching the same value in this file. We should've
                        # already fixed this with fix-rules.py's duplicate_subkeys.
                        msg = "File {0} has duplicated key {1}: {2} vs {3}"
                        msg = msg.format(file_path, key, our_line, this_line)
                        raise ValueError(msg)
                    our_line = this_line
            assert our_line
            subkey_mapping[key] = our_line

        # Now we'll remove all the section's subkeys and start over. Include
        # section header but not any of the keys (or potential blank lines
        # in the interior -- but we preserve them on either end of the
        # section).
        prefix = new_contents[:section_range.start+start_offset]
        contents = list(map(lambda key: subkey_mapping[key], subkeys))
        suffix = new_contents[section_range.end+1-end_offset:]

        new_contents = prefix + contents + suffix

    return new_contents



from __future__ import absolute_import
from __future__ import print_function

import os
import sys
from collections import namedtuple

from .shims import subprocess_check_output, Queue
from .xccdf import get_profile_choices_for_input, get_profile_short_id
from .xccdf import PROFILE_ID_SKIPLIST
from .constants import OSCAP_DS_STRING, OSCAP_PATH


def get_path_args(args):
    """
    Return a namedtuple of (input_path, input_basename, path_base,
    output_dir) from an argparse containing args.input and args.output.
    """

    paths = namedtuple('paths', ['input_path', 'input_basename',
                                 'path_base', 'output_dir'])

    input_path = os.path.abspath(args.input)

    output_dir = os.path.abspath(args.output)
    input_basename = os.path.basename(input_path)

    path_base, _ = os.path.splitext(input_basename)
    # avoid -ds and -xccdf suffices in guide filenames
    if path_base.endswith("-ds"):
        path_base = path_base[:-3]
    elif path_base.endswith("-xccdf"):
        path_base = path_base[:-6]

    return paths(input_path, input_basename, path_base, output_dir)


def generate_for_input_content(input_content, benchmark_id, profile_id):
    """
    Returns HTML guide for given input_content and profile_id
    combination. This function assumes only one Benchmark exists
    in given input_content!
    """

    args = [OSCAP_PATH, "xccdf", "generate", "guide"]
    if benchmark_id != "":
        args.extend(["--benchmark-id", benchmark_id])
    if profile_id != "":
        args.extend(["--profile", profile_id])
    args.append(input_content)

    return subprocess_check_output(args).decode("utf-8")


def builder(queue):
    """
    Fetch from a queue of tasks, process tasks until the queue is empty.
    Each task is processed with generate_for_input_content, and the
    guide is written as output.

    Raises: when an error occurred when processing a task.
    """
    while True:
        try:
            benchmark_id, profile_id, input_path, guide_path = \
                queue.get(False)

            guide_html = generate_for_input_content(
                input_path, benchmark_id, profile_id
            )

            with open(guide_path, "wb") as guide_file:
                guide_file.write(guide_html.encode("utf-8"))

            queue.task_done()
        except Queue.Empty:
            break
        except Exception as error:
            sys.stderr.write(
                "Fatal error encountered when generating guide '%s'. "
                "Error details:\n%s\n\n" % (guide_path, error)
            )
            with queue.mutex:
                queue.queue.clear()
            raise error


def _benchmark_profile_pair_sort_key(benchmark_id, profile_id, profile_title):
    # The "base" benchmarks come first
    if (benchmark_id.endswith("_RHEL-7") or
            benchmark_id.endswith("_RHEL-6") or
            benchmark_id.endswith("_RHEL-5")):
        benchmark_id = "AAA" + benchmark_id

    # The default profile comes last
    if not profile_id:
        profile_title = "zzz(default)"

    return (benchmark_id, profile_title)


def get_benchmark_profile_pairs(input_tree, benchmarks):
    benchmark_profile_pairs = []

    for benchmark_id in benchmarks.keys():
        profiles = get_profile_choices_for_input(input_tree, benchmark_id,
                                                 None)
        for profile_id in profiles:
            pair = (benchmark_id, profile_id, profiles[profile_id])
            benchmark_profile_pairs.append(pair)

    return sorted(benchmark_profile_pairs, key=lambda x:
                  _benchmark_profile_pair_sort_key(x[0], x[1], x[2]))


def _is_skipped_profile(profile_id):
    for skipped_id in PROFILE_ID_SKIPLIST:
        if profile_id.endswith(skipped_id):
            return True
    return False


def _get_guide_filename(path_base, profile_id, benchmark_id, benchmarks):
    profile_id_for_path = "default" if not profile_id else profile_id
    benchmark_id_for_path = benchmark_id
    if benchmark_id_for_path.startswith(OSCAP_DS_STRING):
        benchmark_id_for_path = \
            benchmark_id_for_path[len(OSCAP_DS_STRING):]

    if len(benchmarks) == 1 or len(benchmark_id_for_path) == len("RHEL-X"):
        # treat the base RHEL benchmark as a special case to preserve
        # old guide paths and old URLs that people may be relying on
        return "%s-guide-%s.html" % (path_base,
                                     get_profile_short_id(profile_id_for_path))

    return "%s-%s-guide-%s.html" % \
        (path_base, benchmark_id_for_path,
         get_profile_short_id(profile_id_for_path))


def get_output_guide_paths(benchmarks, benchmark_profile_pairs, path_base,
                           output_dir):
    """
    Return a list of guide paths containing guides for each non-skipped
    profile_id in a benchmark.
    """

    guide_paths = []

    for benchmark_id, profile_id, _ in benchmark_profile_pairs:
        if _is_skipped_profile(profile_id):
            continue

        guide_filename = _get_guide_filename(path_base, profile_id,
                                             benchmark_id, benchmarks)
        guide_path = os.path.join(output_dir, guide_filename)

        guide_paths.append(guide_path)

    return guide_paths


def fill_queue(benchmarks, benchmark_profile_pairs, input_path, path_base,
               output_dir):
    """
    For each benchmark and profile in the benchmark, create a queue of
    tasks for later processing. A task is a named tuple (benchmark_id,
    profile_id, input_path, guide_path).

    Returns: queue of tasks.
    """

    index_links = []
    index_options = {}
    index_initial_src = None
    queue = Queue.Queue()

    task = namedtuple('task', ['benchmark_id', 'profile_id', 'input_path', 'guide_path'])

    for benchmark_id, profile_id, profile_title in benchmark_profile_pairs:
        if _is_skipped_profile(profile_id):
            continue

        guide_filename = _get_guide_filename(path_base, profile_id,
                                             benchmark_id, benchmarks)
        guide_path = os.path.join(output_dir, guide_filename)

        index_links.append(
            "<a target=\"guide\" href=\"%s\">%s</a>" %
            (guide_filename, "%s in %s" % (profile_title, benchmark_id))
        )

        if benchmark_id not in index_options:
            index_options[benchmark_id] = []

        index_options[benchmark_id].append(
            "<option value=\"%s\" data-benchmark-id=\"%s\" data-profile-id=\"%s\">%s</option>" %
            (guide_filename,
             "" if len(benchmarks) == 1 else benchmark_id, profile_id,
             profile_title)
        )

        if index_initial_src is None:
            index_initial_src = guide_filename

        queue.put(task(benchmark_id, profile_id, input_path, guide_path))

    return index_links, index_options, index_initial_src, queue


def build_index(benchmarks, input_basename, index_links, index_options,
                index_initial_src):
    index_select_options = ""
    if len(index_options.keys()) > 1:
        # we sort by length of the benchmark_id to make sure the "default"
        # comes up first in the list
        for benchmark_id in sorted(index_options.keys(),
                                   key=lambda val: (len(val), val)):
            index_select_options += "<optgroup label=\"benchmark: %s\">\n" \
                % (benchmark_id)
            index_select_options += "\n".join(index_options[benchmark_id])
            index_select_options += "</optgroup>\n"
    else:
        index_select_options += "\n".join(list(index_options.values())[0])

    return "".join([
        "<!DOCTYPE html>\n",
        "<html lang=\"en\">\n",
        "\t<head>\n",
        "\t\t<meta charset=\"utf-8\">\n",
        "\t\t<title>%s</title>\n" % (list(benchmarks.values())[0]),
        "\t\t<script>\n",
        "\t\t\tfunction change_profile(option_element)\n",
        "\t\t\t{\n",
        "\t\t\t\tvar benchmark_id=option_element.getAttribute('data-benchmark-id');\n",
        "\t\t\t\tvar profile_id=option_element.getAttribute('data-profile-id');\n",
        "\t\t\t\tvar eval_snippet=document.getElementById('eval_snippet');\n",
        "\t\t\t\tvar input_path='/usr/share/xml/scap/ssg/content/%s';\n" % (input_basename),
        "\t\t\t\tif (profile_id == '')\n",
        "\t\t\t\t{\n",
        "\t\t\t\t\tif (benchmark_id == '')\n",
        "\t\t\t\t\t\teval_snippet.innerHTML='# oscap xccdf eval ' + input_path;\n",
        "\t\t\t\t\telse\n",
        "\t\t\t\t\t\teval_snippet.innerHTML='# oscap xccdf eval --benchmark-id ' + benchmark_id + ' &#92;<br/>' + input_path;\n",
        "\t\t\t\t}\n",
        "\t\t\t\telse\n",
        "\t\t\t\t{\n",
        "\t\t\t\t\tif (benchmark_id == '')\n",
        "\t\t\t\t\t\teval_snippet.innerHTML='# oscap xccdf eval --profile ' + profile_id + ' &#92;<br/>' + input_path;\n",
        "\t\t\t\t\telse\n",
        "\t\t\t\t\t\teval_snippet.innerHTML='# oscap xccdf eval --benchmark-id ' + benchmark_id + ' &#92;<br/>--profile ' + profile_id + ' &#92;<br/>' + input_path;\n",
        "\t\t\t\t}\n",
        "\t\t\t\twindow.open(option_element.value, 'guide');\n",
        "\t\t\t}\n",
        "\t\t</script>\n",
        "\t\t<style>\n",
        "\t\t\thtml, body { margin: 0; height: 100% }\n",
        "\t\t\t#js_switcher { position: fixed; right: 30px; top: 10px; padding: 2px; background: #ddd; border: 1px solid #999 }\n",
        "\t\t\t#guide_div { margin: auto; width: 99%; height: 99% }\n",
        "\t\t</style>\n",
        "\t</head>\n",
        "\t<body onload=\"document.getElementById('js_switcher').style.display = 'block'\">\n",
        "\t\t<noscript>\n",
        "Profiles: ",
        ", ".join(index_links) + "\n",
        "\t\t</noscript>\n",
        "\t\t<div id=\"js_switcher\" style=\"display: none\">\n",
        "\t\t\tProfile: \n",
        "\t\t\t<select style=\"margin-bottom: 5px\" ",
        "onchange=\"change_profile(this.options[this.selectedIndex]);\"",
        ">\n",
        "\n", index_select_options, "\n",
        "\t\t\t</select>\n",
        "\t\t\t<div id='eval_snippet' style='background: #eee; padding: 3px; border: 1px solid #000'>",
        "select a profile to display its guide and a command line snippet needed to use it",
        "</div>\n",
        "\t\t</div>\n",
        "\t\t<div id=\"guide_div\">\n",
        "\t\t\t<iframe src=\"%s\" name=\"guide\" " % (index_initial_src),
        "width=\"100%\" height=\"100%\">\n",
        "\t\t\t</iframe>\n",
        "\t\t</div>\n",
        "\t</body>\n",
        "</html>\n"
    ])



"""
Common functions for OVAL parsing in SSG
"""

from __future__ import absolute_import
from __future__ import print_function

from .xml import ElementTree as ET


REFERENCE_TO_GROUP = dict(
    var_ref="variables",
    test_ref="tests",
    object_ref="objects",
    state_ref="states",
)


CONTAINER_GROUPS = set((
    "definitions",
    "objects",
    "states",
    "tests",
    "variables",
))


class ElementFinder(object):
    """
    A class to find specific elements within an XML structure based on given criteria.

    Attributes:
        oval_groups (dict): A dictionary containing groups of OVAL definitions.
        target (str): The name of the target element to find.
        attrib (str): The attribute of the target element to retrieve.
        result (set): A set to store the results of the found elements' attributes.
    """
    def __init__(self, oval_groups):
        self.oval_groups = oval_groups
        self.target = None
        self.attrib = None
        self.result = set()

    def find_element(self, start_element, target_element_name, sought_attrib):
        """
        Find elements in an XML tree that match the target element name and sought attribute.

        Args:
            start_element (Element): The starting element to begin the search from.
            target_element_name (str): The name of the target element to find.
            sought_attrib (str): The attribute to look for in the target elements.

        Returns:
            None: This method does not return a value. The results are stored in the instance
                  variable `self.result`.
        """
        self.target = target_element_name
        self.attrib = sought_attrib
        self.result = set()
        self._recurse(start_element)

    def _recurse(self, element):
        """
        Recursively traverses an XML element tree to find elements with a specific tag and attribute.

        Args:
            element (xml.etree.ElementTree.Element): The current XML element to examine.

        Returns:
            None
        """
        if element.tag.endswith(self.target):
            self.result.add(element.attrib[self.attrib])
            return

        self._examine_element(element)
        for child in element:
            self._recurse(child)

    def _examine_element(self, element):
        """
        Examines an XML element and processes it based on its tag name or reference attributes.

        This method first strips the namespace from the element's tag name and checks if it is in
        the REFERENCE_TO_GROUP dictionary. If it is, it retrieves the corresponding group from the
        oval_groups dictionary using the element's text as the key. If the tag name is not in
        REFERENCE_TO_GROUP, it searches for reference attributes within the element. If reference
        attributes are found, it retrieves the corresponding group from the oval_groups dictionary
        using the reference attribute name and entity ID.

        If a new root element is found, the method recursively processes the new root element.

        Args:
            element (xml.etree.ElementTree.Element): The XML element to examine and process.

        Raises:
            AssertionError: If the entity ID is not found in the corresponding group in oval_groups.
        """
        name = _strip_ns_from_tag(element.tag)
        new_root = None

        if name in REFERENCE_TO_GROUP:
            reference_target = REFERENCE_TO_GROUP[name]
            new_root = self.oval_groups[reference_target][element.text]
        else:
            _attr_group = _search_element_for_reference_attributes(element)
            if _attr_group is not None:
                ref_attribute_name, entity_id = _attr_group
                reference_target = REFERENCE_TO_GROUP[ref_attribute_name]
                assert entity_id in self.oval_groups[reference_target], \
                    ('Missing definition: "%s" in "%s" "%s"' %
                     (entity_id, reference_target, element))
                new_root = self.oval_groups[reference_target][entity_id]

        if new_root is not None:
            self._recurse(new_root)


def _sort_by_id(elements):
    """
    Process a list of XML elements by their 'id' attribute and returns a dictionary.

    Args:
        elements (list): A list of XML elements, each having an 'id' attribute.

    Returns:
        dict: A dictionary where the keys are the 'id' attributes of the elements and the values
              are the corresponding XML elements.
    """
    ret = dict()
    for element in elements:
        ret[element.attrib["id"]] = element
    return ret


def _search_dict_for_items_that_end_with(dic, what_to_look_for):
    """
    Searches a dictionary for keys that end with a specified substring and returns the
    corresponding value.

    Args:
        dic (dict): The dictionary to search through.
        what_to_look_for (str): The substring to look for at the end of the dictionary keys.

    Returns:
        The value corresponding to the first key that ends with the specified substring,
        or None if no such key is found.
    """
    for item in dic:
        if item.endswith(what_to_look_for):
            return dic[item]
    return None


def _search_element_for_reference_attributes(element):
    """
    Searches an XML element for attributes that match reference names defined in REFERENCE_TO_GROUP.

    Args:
        element (xml.etree.ElementTree.Element): The XML element to search within.

    Returns:
        tuple: A tuple containing the reference attribute name and its occurrence if found. Or
        None, if no matching reference attribute is found.
    """
    for ref_attribute_name in REFERENCE_TO_GROUP:
        occurence = _search_dict_for_items_that_end_with(
            element.attrib, ref_attribute_name)
        if occurence is not None:
            return ref_attribute_name, occurence
    return None


def _find_attr(oval_groups, defn, elem, attr):
    """
    Finds and returns the attribute value of a specified element within a given definition.

    Args:
        oval_groups (list): A list of OVAL groups to search within.
        defn (str): The definition identifier to search for.
        elem (str): The element name to find within the definition.
        attr (str): The attribute name whose value is to be retrieved.

    Returns:
        str: The value of the specified attribute if found, otherwise None.
    """
    finder = ElementFinder(oval_groups)
    finder.find_element(defn, elem, attr)
    return finder.result


def resolve_definition(oval_groups, defn):
    """
    Resolves a definition by finding the attribute 'external_variable' with the specified 'id'
    in the given OVAL groups.

    Args:
        oval_groups (dict): A dictionary containing OVAL groups.
        defn (str): The definition to resolve.

    Returns:
        The value of the 'external_variable' attribute with the specified 'id' if found,
        otherwise None.
    """
    return _find_attr(oval_groups, defn, "external_variable", "id")


def find_extending_defs(oval_groups, defn):
    """
    Find and return the definitions that extend a given definition.

    Args:
        oval_groups (dict): A dictionary containing OVAL groups.
        defn (str): The definition to find extensions for.

    Returns:
        list: A list of definitions that extend the given definition.
    """
    return _find_attr(oval_groups, defn, "extend_definition", "definition_ref")


def get_container_groups(fname):
    """
    Parses an OVAL file and retrieves container groups.

    Args:
        fname (str): The path to the OVAL file to be parsed.

    Returns:
        list: A list of container groups extracted from the OVAL file.
    """
    return _get_container_oval_groups_from_tree(ET.parse(fname))


def _strip_ns_from_tag(tag_name):
    """
    Remove the namespace from an XML tag.

    Args:
        tag_name (str): The XML tag with namespace.

    Returns:
        str: The XML tag without the namespace.
    """
    return tag_name.split("}", 1)[1]


def _get_container_oval_groups_from_tree(element_tree):
    """
    Extracts and sorts OVAL groups from an XML element tree.

    This function takes an XML element tree, retrieves its root, and iterates through its
    children. For each child, it strips the namespace from the tag and checks if the resulting
    group name is in the predefined list of container groups (CONTAINER_GROUPS). If it is, the
    child is sorted by ID and added to the oval_groups dictionary under the corresponding group
    name.

    Args:
        element_tree (ElementTree): The XML element tree to parse.

    Returns:
        dict: A dictionary where the keys are group names and the values are the sorted elements
              of those groups.
    """
    root = element_tree.getroot()

    oval_groups = {}
    for child in root:
        group_name = _strip_ns_from_tag(child.tag)
        if group_name in CONTAINER_GROUPS:
            oval_groups[group_name] = _sort_by_id(child)

    return oval_groups


def _get_resolved_definitions(oval_groups):
    """
    Resolves definitions within the given OVAL groups and maps each definition ID to its
    corresponding variable IDs.

    Args:
        oval_groups (dict): A dictionary containing OVAL groups, where "definitions" is a key that
                            maps to another dictionary of definition IDs and their corresponding
                            elements.

    Returns:
        dict: A dictionary mapping each definition ID to a list of variable IDs resolved from the
              definitions.
    """
    def_id_to_vars_ids = {}
    for def_id, def_el in oval_groups["definitions"].items():
        def_id_to_vars_ids[def_id] = resolve_definition(oval_groups, def_el)
    return def_id_to_vars_ids


def _check_sanity(oval_groups, resolved_defns):
    """
    Checks the sanity of OVAL variable definitions by ensuring that all external variables are
    accounted for and no unexpected variables are caught.

    Args:
        oval_groups (dict): A dictionary containing OVAL variable groups. The key "variables"
                            should map to another dictionary where keys are variable IDs and
                            values are variable elements.
        resolved_defns (dict): A dictionary where keys are variable IDs and values are sets of
                               resolved variable definitions.
    Returns:
        None

    Raises:
        AssertionError: If there are unexpected caught variables.
    """
    all_external_variables = set()
    for var_id, var_el in oval_groups["variables"].items():
        if var_el.tag.endswith("external_variable"):
            all_external_variables.add(var_id)

    all_caught_variables = set()
    for var in resolved_defns.values():
        all_caught_variables.update(var)

    skipped_variables = all_external_variables.difference(all_caught_variables)
    if skipped_variables:
        print("These variables managed to slip past:", skipped_variables)
        strange_variables = all_caught_variables.difference(
            all_external_variables)
        assert not strange_variables, \
            ("There were unexpected caught variables: {}"
             .format(str(strange_variables)))


def _check_sanity_on_file(fname):
    """
    Perform sanity checks on the given file.

    This function reads the file specified by `fname`, extracts OVAL groups,
    resolves definitions, and then checks the sanity of these groups and definitions.

    Args:
        fname (str): The path to the file to be checked.

    Returns:
        None

    Raises:
        ValueError: If any sanity check fails.
    """
    oval_groups = get_container_groups(fname)
    resolved_defns = _get_resolved_definitions(oval_groups)
    _check_sanity(oval_groups, resolved_defns)



#!/usr/bin/python3

import argparse
import sys

from ssg.content_diff import StandardContentDiffer, StigContentDiffer
from ssg.xml import XMLContent
import xml.etree.ElementTree as ET


def parse_args():
    parser = argparse.ArgumentParser(
        description="Compares two data streams with regards to presence of"
        "OVAL checks and all remediations")
    parser.add_argument(
        "old", metavar="OLD_DS_PATH",
        help="Path to the old data stream")
    parser.add_argument(
        "new", metavar="NEW_DS_PATH",
        help="Path to the new data stream")
    parser.add_argument(
        "--rule", metavar="RULE_ID",
        help="Compare only the rule specified by given RULE_ID"
    )
    parser.add_argument(
        "--no-diffs", action="store_true",
        help="Do not perform detailed comparison of checks and remediations contents."
    )
    parser.add_argument(
        "--only-rules", action="store_true",
        help="Print only removals from rule set."
    )
    parser.add_argument(
        "--rule-diffs", action="store_true",
        help="Output diffs per rule, instead of a single diff. "
             "The rule diffs are output to directory './compare_ds-diffs/'."
    )
    parser.add_argument(
        "--output-dir", metavar="OUTPUT_DIR",
        type=str, action="store", default="./compare_ds-diffs",
        help="Directory where rule diff files will be saved. Only used with --rule-diffs option. "
             "If the directory doesn't exist, it will be created."
    )
    parser.add_argument(
        "--disa-content", action="store_true",
        help="This option enables comparison between DISA contents by ignoring the release "
             "number in rule IDs."
    )
    return parser.parse_args()


def main():
    args = parse_args()
    old_tree = ET.parse(args.old)
    new_tree = ET.parse(args.new)
    old_root = old_tree.getroot()
    new_root = new_tree.getroot()

    old_xml_content = XMLContent(old_root)
    new_xml_content = XMLContent(new_root)

    if args.disa_content:
        content_differ = StigContentDiffer(old_xml_content, new_xml_content, args.rule,
                                           not args.no_diffs, args.rule_diffs,
                                           args.only_rules, args.output_dir)
    else:
        content_differ = StandardContentDiffer(old_xml_content, new_xml_content, args.rule,
                                               not args.no_diffs, args.rule_diffs,
                                               args.only_rules, args.output_dir)

    for old_benchmark in old_xml_content.get_benchmarks():
        old_benchmark_id = old_benchmark.get_attr("id")
        new_benchmark = new_xml_content.find_benchmark(old_benchmark_id)
        if not new_benchmark:
            print(
                "Warning: Skipping comparison of the following benchmark "
                "because it was not found in the new data stream: {}".format(old_benchmark_id))
            continue

        content_differ.compare_rules(old_benchmark, new_benchmark)
    return 0


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/python3

from __future__ import print_function

import argparse
import glob
import os.path
import sys

import ssg.products
import ssg.yaml

from tests.common import stability


IGNORED_PROPERTIES = (
    "product_dir",
)


def describe_modification(intro, changeset):
    if not changeset:
        return ""

    msg = intro
    for what, (initial, final) in changeset.items():
        msg += " - {what} from '{initial}' -> '{final}'\n".format(
                what=what, initial=initial, final=final)
    return msg


def describe_change(difference, name):
    msg = ""

    msg += stability.describe_changeset(
        "Following properties were added to the {name} product:\n".format(name=name),
        difference.added,
    )
    msg += stability.describe_changeset(
        "Following properties were removed from the {name} product:\n".format(name=name),
        difference.removed,
    )
    msg += describe_modification(
        "Following properties got different values in the {name} product:\n".format(name=name),
        difference.modified,
    )
    return msg.rstrip()


def compare_dictionaries(reference, sample):
    reference_keys = set(reference.keys())
    sample_keys = set(sample.keys())

    result = stability.Difference()
    result.added = list(sample_keys.difference(reference_keys))
    result.removed = list(reference_keys.difference(sample_keys))
    for key, value in reference.items():
        if sample.get(key, value) != value:
            result.modified[key] = (value, sample[key])
    for item in IGNORED_PROPERTIES:
        result.remove_item_from_comparison(item)
    return result


def get_references_filenames(ref_root):
    return glob.glob(os.path.join(ref_root, "*.yml"))


def corresponding_product_built(build_dir, product_id):
    return os.path.isdir(os.path.join(build_dir, product_id))


def get_matching_compiled_product_filename(build_dir, product_id):
    ref_path_components = reference_fname.split(os.path.sep)
    matching_filename = os.path.join(build_dir, product_id, "product.yml")
    if os.path.isfile(matching_filename):
        return matching_filename


def get_reference_vs_built_difference(ref_product, built_product):
    ref_dict = dict()
    ref_dict.update(ref_product)
    built_dict = dict()
    built_dict.update(built_product)
    difference = compare_dictionaries(ref_dict, built_dict)
    return difference


def inform_and_append_fix_based_on_reference_compiled_product(ref, build_root):
    ref_product = ssg.products.Product(ref)
    product_id = ref_product["product"]
    if not corresponding_product_built(build_root, product_id):
        return True

    compiled_path = os.path.join(build_root, product_id, "product.yml")
    compiled_product = ssg.products.Product(compiled_path)
    difference = get_reference_vs_built_difference(ref_product, compiled_product)
    all_ok = difference.empty
    return all_ok


def compare_compiled_products_with_reference_data(test_data_root, build_root):
    reference_files = get_references_filenames(test_data_root)
    if not reference_files:
        raise RuntimeError("Unable to find any reference compiled products in {test_root}"
                           .format(test_root=test_data_root))
    all_ok = True
    for ref in reference_files:
        all_ok &= inform_and_append_fix_based_on_reference_compiled_product(
                ref, build_root)

    if not all_ok:
        products_dir = os.path.join(os.path.dirname(__file__), "..", "products")
        msg = (
            "If changes to mentioned products are intentional, "
            "execute this script with PYTHONPATH set to the project root "
            "to regenerate reference products: \n"
            "{script_name} --update-reference-data {build_root} {test_data_root}\n"
            "If those changes are unwanted, take a look at product properties "
            "that likely cause these changes."
            .format(script_name=__file__, build_root=os.path.normpath(products_dir),
                    test_data_root=test_data_root))
        print(msg, file=sys.stderr)
    return all_ok


def write_product_without_keys(product, path, exclude_keys=frozenset()):
    product_data = dict()
    product_data.update(product)
    for key in exclude_keys:
        product_data.pop(key, None)
    with open(path, "w") as f:
        ssg.yaml.ordered_dump(product_data, f)


def update_reference_data(test_data_root, products_dir):
    properties_dir = os.path.join(products_dir, os.path.pardir, "product_properties")
    product_yamls = glob.glob(os.path.join(products_dir, "*", "product.yml"))
    for fname in product_yamls:
        product = ssg.products.Product(fname)
        product.read_properties_from_directory(properties_dir)
        out_fname = os.path.join(test_data_root, product["product"] + ".yml")
        write_product_without_keys(product, out_fname, IGNORED_PROPERTIES)
    print("Updated test data of {num_products} products".format(num_products=len(product_yamls)))


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
            "build_root", help="A directory that contains product-named subdirectories "
            "with product.yml files. Typically the build root, or products source root "
            "if the script is executed in the update mode")
    parser.add_argument("test_data_root", help="Root of reference data containing "
                        "reference files named <product>.yml")
    parser.add_argument(
        "--update-reference-data", action="store_true", default=False,
        help="If supplied with this option, "
        "and the build root is the product root, update all of the test data")
    args = parser.parse_args()

    if args.update_reference_data:
        update_reference_data(args.test_data_root, args.build_root)
        sys.exit(0)

    all_ok = compare_compiled_products_with_reference_data(args.test_data_root, args.build_root)
    return_code = int(not all_ok)
    sys.exit(return_code)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

# Usage (Execute in root project directory): python3 -m pytest -n auto tests/test_thin_ds.py

import pytest
import subprocess

from lxml import etree

from pathlib import Path


THIN_DS_PATHS = []
for path in Path("./build/thin_ds").glob("*.xml"):
    THIN_DS_PATHS.append(path)


@pytest.mark.parametrize("path", THIN_DS_PATHS)
def test_thin_ds(path):
    root = etree.parse(path)

    rules = root.findall(".//{http://checklists.nist.gov/xccdf/1.2}Rule")
    assert len(rules) == 1

    rule_id = rules[0].get("id", "NoID").removeprefix("xccdf_org.ssgproject.content_rule_")
    assert rule_id in str(path)

    profiles = root.findall(".//{http://checklists.nist.gov/xccdf/1.2}Profile")
    assert len(profiles) == 1

    profile_id = (
        profiles[0].get("id", "NoID").removeprefix("xccdf_org.ssgproject.content_profile_")
    )
    assert profile_id in str(path)


@pytest.mark.parametrize("path", THIN_DS_PATHS)
def test_validate_thin_ds(path):
    command = subprocess.run(["oscap", "ds", "sds-validate", path], capture_output=True)
    assert command.returncode == 0



#!/usr/bin/python3

import argparse

import oval_tester


def main():
    parser = argparse.ArgumentParser(
        description="Test Jinja macros that Generate OVAL")
    parser.add_argument(
        "--verbose", action="store_true", default=False,
        help="Show results of each test case")
    args = parser.parse_args()
    tester = oval_tester.OVALTester(args.verbose)

    #######################################################
    # Test cases for whitespace separated files
    #######################################################

    tester.test(
        "correct value",
        r"""{{{ oval_check_config_file(
            path='CONFIG_FILE',
            prefix_regex='^[ \t]*',
            parameter='speed',
            separator_regex='[ \t]+',
            value='100',
            missing_parameter_pass=false,
            application='',
            multi_value=false,
            missing_config_file_fail=false,
            section=''
        ) }}}""",
        "speed 100",
        "true"
    )
    tester.test(
        "correct value and a comment",
        r"""{{{ oval_check_config_file(
            path='CONFIG_FILE',
            prefix_regex='^[ \t]*',
            parameter='speed',
            separator_regex='[ \t]+',
            value='100',
            missing_parameter_pass=false,
            application='',
            multi_value=false,
            missing_config_file_fail=false,
            section=''
        ) }}}""",
        "speed 100 # be very fast",
        "true"
    )
    tester.test(
        "correct value on a new line",
        r"""{{{ oval_check_config_file(
            path='CONFIG_FILE',
            prefix_regex='^[ \t]*',
            parameter='speed',
            separator_regex='[ \t]+',
            value='100',
            missing_parameter_pass=false,
            application='',
            multi_value=false,
            missing_config_file_fail=false,
            section=''
        ) }}}""",
        "\n\n\n\n\n\nspeed 100",
        "true"
    )
    tester.test(
        "correct value separated by a tab",
        r"""{{{ oval_check_config_file(
            path='CONFIG_FILE',
            prefix_regex='^[ \t]*',
            parameter='speed',
            separator_regex='[ \t]+',
            value='100',
            missing_parameter_pass=false,
            application='',
            multi_value=false,
            missing_config_file_fail=false,
            section=''
        ) }}}""",
        "speed\t100",
        "true"
    )
    tester.test(
        "wrong value",
        r"""{{{ oval_check_config_file(
            path='CONFIG_FILE',
            prefix_regex='^[ \t]*',
            parameter='speed',
            separator_regex='[ \t]+',
            value='100',
            missing_parameter_pass=false,
            application='',
            multi_value=false,
            missing_config_file_fail=false,
            section=''
        ) }}}""",
        "speed 80",
        "false"
    )
    tester.test(
        "wrong value which contains the correct value as a substring",
        r"""{{{ oval_check_config_file(
            path='CONFIG_FILE',
            prefix_regex='^[ \t]*',
            parameter='speed',
            separator_regex='[ \t]+',
            value='100',
            missing_parameter_pass=false,
            application='',
            multi_value=false,
            missing_config_file_fail=false,
            section=''
        ) }}}""",
        "speed 1000",
        "false"
    )
    tester.test(
        "commented value",
        r"""{{{ oval_check_config_file(
            path='CONFIG_FILE',
            prefix_regex='^[ \t]*',
            parameter='speed',
            separator_regex='[ \t]+',
            value='100',
            missing_parameter_pass=false,
            application='',
            multi_value=false,
            missing_config_file_fail=false,
            section=''
        ) }}}""",
        "# speed 80",
        "false"
    )
    tester.test(
        "missing whitespace",
        r"""{{{ oval_check_config_file(
            path='CONFIG_FILE',
            prefix_regex='^[ \t]*',
            parameter='speed',
            separator_regex='[ \t]+',
            value='100',
            missing_parameter_pass=false,
            application='',
            multi_value=false,
            missing_config_file_fail=false,
            section=''
        ) }}}""",
        "speed100",
        "false"
    )
    tester.test(
        "parameter without a value",
        r"""{{{ oval_check_config_file(
            path='CONFIG_FILE',
            prefix_regex='^[ \t]*',
            parameter='speed',
            separator_regex='[ \t]+',
            value='100',
            missing_parameter_pass=false,
            application='',
            multi_value=false,
            missing_config_file_fail=false,
            section=''
        ) }}}""",
        "speed ",
        "false"
    )
    tester.test(
        "misspelled parameter with a value",
        r"""{{{ oval_check_config_file(
            path='CONFIG_FILE',
            prefix_regex='^[ \t]*',
            parameter='speed',
            separator_regex='[ \t]+',
            value='100',
            missing_parameter_pass=false,
            application='',
            multi_value=false,
            missing_config_file_fail=false,
            section=''
        ) }}}""",
        "sspeed 100",
        "false"
    )
    tester.test(
        "parameter is on a different line than the value",
        r"""{{{ oval_check_config_file(
            path='CONFIG_FILE',
            prefix_regex='^[ \t]*',
            parameter='speed',
            separator_regex='[ \t]+',
            value='100',
            missing_parameter_pass=false,
            application='',
            multi_value=false,
            missing_config_file_fail=false,
            section=''
        ) }}}""",
        "speed\n100",
        "false"
    )
    tester.test(
        "multiple empty lines among parameter and value",
        r"""{{{ oval_check_config_file(
            path='CONFIG_FILE',
            prefix_regex='^[ \t]*',
            parameter='speed',
            separator_regex='[ \t]+',
            value='100',
            missing_parameter_pass=false,
            application='',
            multi_value=false,
            missing_config_file_fail=false,
            section=''
        ) }}}""",
        "\n\nspeed\n\n\n100\n\n\n\n",
        "false"
    )
    tester.test(
        "parameter with multiple values when multi_value disabled",
        r"""{{{ oval_check_config_file(
            path='CONFIG_FILE',
            prefix_regex='^[ \t]*',
            parameter='speed',
            separator_regex='[ \t]+',
            value='100',
            missing_parameter_pass=false,
            application='',
            multi_value=false,
            missing_config_file_fail=false,
            section=''
        ) }}}""",
        "speed 100 50 8",
        "false"
    )
    tester.test(
        "parameter with single value when multi_value enabled",
        r"""{{{ oval_check_config_file(
            path='CONFIG_FILE',
            prefix_regex='^[ \t]*',
            parameter='speed',
            separator_regex='[ \t]+',
            value='100',
            missing_parameter_pass=false,
            application='',
            multi_value=true,
            missing_config_file_fail=false,
            section=''
        ) }}}""",
        "speed 100",
        "true"
    )
    tester.test(
        "parameter, single value, trailing comment, multi_value enabled",
        r"""{{{ oval_check_config_file(
            path='CONFIG_FILE',
            prefix_regex='^[ \t]*',
            parameter='speed',
            separator_regex='[ \t]+',
            value='100',
            missing_parameter_pass=false,
            application='',
            multi_value=true,
            missing_config_file_fail=false,
            section=''
        ) }}}""",
        "speed 100 #comment",
        "true"
    )
    tester.test(
        "parameter with multiple values when multi_value enabled",
        r"""{{{ oval_check_config_file(
            path='CONFIG_FILE',
            prefix_regex='^[ \t]*',
            parameter='speed',
            separator_regex='[ \t]+',
            value='100',
            missing_parameter_pass=false,
            application='',
            multi_value=true,
            missing_config_file_fail=false,
            section=''
        ) }}}""",
        "speed 100 50 8",
        "true"
    )
    tester.test(
        "multiple values, multi_value enabled, value in the middle",
        r"""{{{ oval_check_config_file(
            path='CONFIG_FILE',
            prefix_regex='^[ \t]*',
            parameter='speed',
            separator_regex='[ \t]+',
            value='100',
            missing_parameter_pass=false,
            application='',
            multi_value=true,
            missing_config_file_fail=false,
            section=''
        ) }}}""",
        "speed abcd 333 100 50 8",
        "true"
    )
    tester.test(
        "multiple values, multi_value enabled, value is the last",
        r"""{{{ oval_check_config_file(
            path='CONFIG_FILE',
            prefix_regex='^[ \t]*',
            parameter='speed',
            separator_regex='[ \t]+',
            value='100',
            missing_parameter_pass=false,
            application='',
            multi_value=true,
            missing_config_file_fail=false,
            section=''
        ) }}}""",
        "speed 2 4 6 8 10 14 100",
        "true"
    )
    tester.test(
        "parameter with extra newlines, multi_value enabled",
        r"""{{{ oval_check_config_file(
            path='CONFIG_FILE',
            prefix_regex='^[ \t]*',
            parameter='speed',
            separator_regex='[ \t]+',
            value='100',
            missing_parameter_pass=false,
            application='',
            multi_value=true,
            missing_config_file_fail=false,
            section=''
        ) }}}""",
        "speed\n\n100",
        "false"
    )
    tester.test(
        "parameter with multiple values when multi_value enabled, comment",
        r"""{{{ oval_check_config_file(
            path='CONFIG_FILE',
            prefix_regex='^[ \t]*',
            parameter='speed',
            separator_regex='[ \t]+',
            value='100',
            missing_parameter_pass=false,
            application='',
            multi_value=true,
            missing_config_file_fail=false,
            section=''
        ) }}}""",
        "speed 2 4 6 8 10 14 100 # astonishing",
        "true"
    )
    tester.test(
        "multi_value is used and value is a suffix of a value",
        r"""{{{ oval_check_config_file(
            path='CONFIG_FILE',
            prefix_regex='^[ \t]*',
            parameter='speed',
            separator_regex='[ \t]+',
            value='100',
            missing_parameter_pass=false,
            application='',
            multi_value=true,
            missing_config_file_fail=false,
            section=''
        ) }}}""",
        "speed 1001000",
        "false"
    )
    tester.test(
        "parameter with a value commented out",
        r"""{{{ oval_check_config_file(
            path='CONFIG_FILE',
            prefix_regex='^[ \t]*',
            parameter='speed',
            separator_regex='[ \t]+',
            value='100',
            missing_parameter_pass=false,
            application='',
            multi_value=false,
            missing_config_file_fail=false,
            section=''
        ) }}}""",
        "speed # 100",
        "false"
    )
    tester.test(
        "missing parameter fails",
        r"""{{{ oval_check_config_file(
            path='CONFIG_FILE',
            prefix_regex='^[ \t]*',
            parameter='speed',
            separator_regex='[ \t]+',
            value='100',
            missing_parameter_pass=false,
            application='',
            multi_value=false,
            missing_config_file_fail=false,
            section=''
        ) }}}""",
        "lights on",
        "false"
    )
    tester.test(
        "missing parameter pass",
        r"""{{{ oval_check_config_file(
            path='CONFIG_FILE',
            prefix_regex='^[ \t]*',
            parameter='speed',
            separator_regex='[ \t]+',
            value='100',
            missing_parameter_pass=true,
            application='',
            multi_value=false,
            missing_config_file_fail=false,
            section=''
        ) }}}""",
        "lights on",
        "true"
    )
    tester.test(
        "overwriting",
        r"""{{{ oval_check_config_file(
            path='CONFIG_FILE',
            prefix_regex='^[ \t]*',
            parameter='speed',
            separator_regex='[ \t]+',
            value='100',
            missing_parameter_pass=true,
            application='',
            multi_value=false,
            missing_config_file_fail=false,
            section=''
        ) }}}""",
        """speed 100
        speed 60""",
        "false"
    )
    tester.test(
        "overwriting commented out",
        r"""{{{ oval_check_config_file(
            path='CONFIG_FILE',
            prefix_regex='^[ \t]*',
            parameter='speed',
            separator_regex='[ \t]+',
            value='100',
            missing_parameter_pass=true,
            application='',
            multi_value=false,
            missing_config_file_fail=false,
            section=''
        ) }}}""",
        """speed 100
        #speed 60""",
        "true"
    )
    tester.test(
        "config file missing should fail",
        r"""{{{ oval_check_config_file(
            path='CONFIG_FILE',
            prefix_regex='^[ \t]*',
            parameter='speed',
            separator_regex='[ \t]+',
            value='100',
            missing_parameter_pass=true,
            application='',
            multi_value=false,
            missing_config_file_fail=true,
            section=''
        ) }}}""",
        None,
        "false"
    )
    tester.test(
        "config file missing should pass",
        r"""{{{ oval_check_config_file(
            path='CONFIG_FILE',
            prefix_regex='^[ \t]*',
            parameter='speed',
            separator_regex='[ \t]+',
            value='100',
            missing_parameter_pass=true,
            application='',
            multi_value=false,
            missing_config_file_fail=false,
            section=''
        ) }}}""",
        None,
        "true"
    )
    tester.test(
        "config file missing should fail due to missing parameter",
        r"""{{{ oval_check_config_file(
            path='CONFIG_FILE',
            prefix_regex='^[ \t]*',
            parameter='speed',
            separator_regex='[ \t]+',
            value='100',
            missing_parameter_pass=false,
            application='',
            multi_value=false,
            missing_config_file_fail=true,
            section=''
        ) }}}""",
        None,
        "false"
    )
    tester.test(
        "config file missing but missing parameter isn't allowed",
        r"""{{{ oval_check_config_file(
            path='CONFIG_FILE',
            prefix_regex='^[ \t]*',
            parameter='speed',
            separator_regex='[ \t]+',
            value='100',
            missing_parameter_pass=false,
            application='',
            multi_value=false,
            missing_config_file_fail=false,
            section=''
        ) }}}""",
        None,
        "false"
    )

    #######################################################
    # Test cases for equal sign separated files
    #######################################################
    tester.test(
        "correct value, no whitespace",
        r"""{{{ oval_check_config_file(
            path='CONFIG_FILE',
            prefix_regex='^[ \t]*',
            parameter='speed',
            separator_regex='[ \t]*=[ \t]*',
            value='100',
            missing_parameter_pass=false,
            application='',
            multi_value=false,
            missing_config_file_fail=false,
            section=''
        ) }}}""",
        "speed=100",
        "true"
    )
    tester.test(
        "correct value, some spaces in the middle",
        r"""{{{ oval_check_config_file(
            path='CONFIG_FILE',
            prefix_regex='^[ \t]*',
            parameter='speed',
            separator_regex='[ \t]*=[ \t]*',
            value='100',
            missing_parameter_pass=false,
            application='',
            multi_value=false,
            missing_config_file_fail=false,
            section=''
        ) }}}""",
        "speed  =  100",
        "true"
    )
    tester.test(
        "correct value, many spaces everywhere",
        r"""{{{ oval_check_config_file(
            path='CONFIG_FILE',
            prefix_regex='^[ \t]*',
            parameter='speed',
            separator_regex='[ \t]*=[ \t]*',
            value='100',
            missing_parameter_pass=false,
            application='',
            multi_value=false,
            missing_config_file_fail=false,
            section=''
        ) }}}""",
        "    speed  =  100      ",
        "true"
    )
    tester.test(
        "correct value, tabs",
        r"""{{{ oval_check_config_file(
            path='CONFIG_FILE',
            prefix_regex='^[ \t]*',
            parameter='speed',
            separator_regex='[ \t]*=[ \t]*',
            value='100',
            missing_parameter_pass=false,
            application='',
            multi_value=false,
            missing_config_file_fail=false,
            section=''
        ) }}}""",
        "\tspeed\t=\t100",
        "true"
    )
    tester.test(
        "correct value, and a comment",
        r"""{{{ oval_check_config_file(
            path='CONFIG_FILE',
            prefix_regex='^[ \t]*',
            parameter='speed',
            separator_regex='[ \t]*=[ \t]*',
            value='100',
            missing_parameter_pass=false,
            application='',
            multi_value=false,
            missing_config_file_fail=false,
            section=''
        ) }}}""",
        "speed=100 # be very fast",
        "true"
    )
    tester.test(
        "wrong value, and a comment",
        r"""{{{ oval_check_config_file(
            path='CONFIG_FILE',
            prefix_regex='^[ \t]*',
            parameter='speed',
            separator_regex='[ \t]*=[ \t]*',
            value='100',
            missing_parameter_pass=false,
            application='',
            multi_value=false,
            missing_config_file_fail=false,
            section=''
        ) }}}""",
        "speed=800 # be extremely fast",
        "false"
    )
    tester.test(
        "no value, and a comment",
        r"""{{{ oval_check_config_file(
            path='CONFIG_FILE',
            prefix_regex='^[ \t]*',
            parameter='speed',
            separator_regex='[ \t]*=[ \t]*',
            value='100',
            missing_parameter_pass=false,
            application='',
            multi_value=false,
            missing_config_file_fail=false,
            section=''
        ) }}}""",
        "speed= # 100",
        "false"
    )

    ######################################
    # Test cases for INI files
    ######################################
    tester.test(
        "INI correct value",
        r"""{{{ oval_check_ini_file(
            path='CONFIG_FILE',
            section="vehicle",
            parameter='speed',
            value='100',
            missing_parameter_pass=false,
            application='',
            multi_value=false,
            missing_config_file_fail=false,
        ) }}}""",
        """[vehicle]
        speed = 100""",
        "true"
    )
    tester.test(
        "INI correct value trailing whitespace",
        r"""{{{ oval_check_ini_file(
            path='CONFIG_FILE',
            section="vehicle",
            parameter='speed',
            value='100',
            missing_parameter_pass=false,
            application='',
            multi_value=false,
            missing_config_file_fail=false,
        ) }}}""",
        """[vehicle]
        speed = 100               """,
        "true"
    )
    tester.test(
        "INI correct value no whitespace",
        r"""{{{ oval_check_ini_file(
            path='CONFIG_FILE',
            section="vehicle",
            parameter='speed',
            value='100',
            missing_parameter_pass=false,
            application='',
            multi_value=false,
            missing_config_file_fail=false,
        ) }}}""",
        "[vehicle]\nspeed=100",
        "true"
    )
    tester.test(
        "INI correct value tabs",
        r"""{{{ oval_check_ini_file(
            path='CONFIG_FILE',
            section="vehicle",
            parameter='speed',
            value='100',
            missing_parameter_pass=false,
            application='',
            multi_value=false,
            missing_config_file_fail=false,
        ) }}}""",
        "[vehicle]\n\tspeed\t=\t100",
        "true"
    )
    tester.test(
        "INI correct value commented out",
        r"""{{{ oval_check_ini_file(
            path='CONFIG_FILE',
            section="vehicle",
            parameter='speed',
            value='100',
            missing_parameter_pass=false,
            application='',
            multi_value=false,
            missing_config_file_fail=false,
        ) }}}""",
        """[vehicle]
        #speed = 100""",
        "false"
    )
    tester.test(
        "INI section commented out",
        r"""{{{ oval_check_ini_file(
            path='CONFIG_FILE',
            section="vehicle",
            parameter='speed',
            value='100',
            missing_parameter_pass=false,
            application='',
            multi_value=false,
            missing_config_file_fail=false,
        ) }}}""",
        "#[vehicle]\nspeed = 100",
        "false"
    )
    tester.test(
        "INI correct value among multiple values",
        r"""{{{ oval_check_ini_file(
            path='CONFIG_FILE',
            section="vehicle",
            parameter='speed',
            value='100',
            missing_parameter_pass=false,
            application='',
            multi_value=false,
            missing_config_file_fail=false,
        ) }}}""",
        """[vehicle]
        color = red
        speed = 100
        doors = 5""",
        "true"
    )
    tester.test(
        "INI correct value among multiple values commented out",
        r"""{{{ oval_check_ini_file(
            path='CONFIG_FILE',
            section="vehicle",
            parameter='speed',
            value='100',
            missing_parameter_pass=false,
            application='',
            multi_value=false,
            missing_config_file_fail=false,
        ) }}}""",
        """[vehicle]
        color = red
        #speed = 100
        doors = 5""",
        "false"
    )
    tester.test(
        "INI wrong value",
        r"""{{{ oval_check_ini_file(
            path='CONFIG_FILE',
            section="vehicle",
            parameter='speed',
            value='100',
            missing_parameter_pass=false,
            application='',
            multi_value=false,
            missing_config_file_fail=false,
        ) }}}""",
        """[vehicle]
        speed = 200""",
        "false"
    )
    tester.test(
        "INI wrong value which is a substring",
        r"""{{{ oval_check_ini_file(
            path='CONFIG_FILE',
            section="vehicle",
            parameter='speed',
            value='100',
            missing_parameter_pass=false,
            application='',
            multi_value=false,
            missing_config_file_fail=false,
        ) }}}""",
        """[vehicle]
        speed = 10000""",
        "false"
    )
    tester.test(
        "INI overwritten",
        r"""{{{ oval_check_ini_file(
            path='CONFIG_FILE',
            section="vehicle",
            parameter='speed',
            value='100',
            missing_parameter_pass=false,
            application='',
            multi_value=false,
            missing_config_file_fail=false,
        ) }}}""",
        """[vehicle]
        speed = 100
        speed = 200""",
        "false"
    )
    tester.test(
        "INI correct value in a wrong section",
        r"""{{{ oval_check_ini_file(
            path='CONFIG_FILE',
            section="vehicle",
            parameter='speed',
            value='100',
            missing_parameter_pass=false,
            application='',
            multi_value=false,
            missing_config_file_fail=false,
        ) }}}""",
        """[house]
        speed = 100""",
        "false"
    )
    tester.test(
        "INI section overwritten",
        r"""{{{ oval_check_ini_file(
            path='CONFIG_FILE',
            section="vehicle",
            parameter='speed',
            value='100',
            missing_parameter_pass=false,
            application='',
            multi_value=false,
            missing_config_file_fail=false,
        ) }}}""",
        "[vehicle]\n[house]\nspeed = 100",
        "false"
    )
    tester.test(
        "INI no section at all",
        r"""{{{ oval_check_ini_file(
            path='CONFIG_FILE',
            section="vehicle",
            parameter='speed',
            value='100',
            missing_parameter_pass=false,
            application='',
            multi_value=false,
            missing_config_file_fail=false,
        ) }}}""",
        "speed = 100",
        "false"
    )
    tester.test(
        "INI extra newlines",
        r"""{{{ oval_check_ini_file(
            path='CONFIG_FILE',
            section="vehicle",
            parameter='speed',
            value='100',
            missing_parameter_pass=false,
            application='',
            multi_value=false,
            missing_config_file_fail=false,
        ) }}}""",
        "[vehicle]\nspeed =\n100",
        "false"
    )
    tester.test(
        "SHELL commented out",
        r"""{{{ oval_check_shell_file(
            path='CONFIG_FILE',
            parameter='SHELL',
            value='/bin/bash',
            missing_parameter_pass=false,
            application='',
            multi_value=false,
            missing_config_file_fail=false,
        ) }}}""",
        "# SHELL=/bin/bash\n",
        "false"
    )
    tester.test(
        "SHELL correct",
        r"""{{{ oval_check_shell_file(
            path='CONFIG_FILE',
            parameter='SHELL',
            value='/bin/bash',
            missing_parameter_pass=false,
            application='',
            multi_value=false,
            missing_config_file_fail=false,
        ) }}}""",
        " SHELL=/bin/bash\n",
        "true"
    )
    tester.test(
        "SHELL single-quoted",
        r"""{{{ oval_check_shell_file(
            path='CONFIG_FILE',
            parameter='SHELL',
            value='/bin"/bash',
            missing_parameter_pass=false,
            application='',
            multi_value=false,
            missing_config_file_fail=false,
        ) }}}""",
        " SHELL='/bin\"/bash'\n",
        "true"
    )
    tester.test(
        "SHELL double-quoted",
        r"""{{{ oval_check_shell_file(
            path='CONFIG_FILE',
            parameter='SHELL',
            value='  /bin/bash',
            missing_parameter_pass=false,
            application='',
            multi_value=false,
            missing_config_file_fail=false,
        ) }}}""",
        """ SHELL="  /bin/bash"\n""",
        "true"
    )
    tester.test(
        "SHELL unwanted double-quoted",
        r"""{{{ oval_check_shell_file(
            path='CONFIG_FILE',
            parameter='SHELL',
            value='  /bin/bash',
            no_quotes=true,
            missing_parameter_pass=false,
            application='',
            multi_value=false,
            missing_config_file_fail=false,
        ) }}}""",
        """ SHELL="  /bin/bash"\n""",
        "false"
    )
    tester.test(
        "SHELL unwanted single-quoted",
        r"""{{{ oval_check_shell_file(
            path='CONFIG_FILE',
            parameter='SHELL',
            value='/bin"/bash',
            no_quotes=true,
            missing_parameter_pass=false,
            application='',
            multi_value=false,
            missing_config_file_fail=false,
        ) }}}""",
        " SHELL='/bin\"/bash'\n",
        "false"
    )
    tester.test(
        "SHELL double-quoted spaced",
        r"""{{{ oval_check_shell_file(
            path='CONFIG_FILE',
            parameter='SHELL',
            value='/bin/bash',
            missing_parameter_pass=false,
            application='',
            multi_value=false,
            missing_config_file_fail=false,
        ) }}}""",
        """ SHELL= "/bin/bash"\n""",
        "false"
    )
    tester.test(
        "SHELL bad_var_case",
        r"""{{{ oval_check_shell_file(
            path='CONFIG_FILE',
            parameter='SHELL',
            value='/bin/bash',
            missing_parameter_pass=false,
            application='',
            multi_value=false,
            missing_config_file_fail=false,
        ) }}}""",
        """ Shell="/bin/bash"\n""",
        "false"
    )
    tester.test(
        "SHELL bad_value_case",
        r"""{{{ oval_check_shell_file(
            path='CONFIG_FILE',
            parameter='SHELL',
            value='/bin/bash',
            missing_parameter_pass=false,
            application='',
            multi_value=false,
            missing_config_file_fail=false,
        ) }}}""",
        """ SHELL="/bin/Bash"\n""",
        "false"
    )
    tester.test(
        "SHELL badly quoted",
        r"""{{{ oval_check_shell_file(
            path='CONFIG_FILE',
            parameter='SHELL',
            value='/bin/bash',
            missing_parameter_pass=false,
            application='',
            multi_value=false,
            missing_config_file_fail=false,
        ) }}}""",
        """ SHELL="/bin/bash'\n""",
        "false"
    )

    tester.finish()


if __name__ == "__main__":
    main()



from ...constants import STR_TO_BOOL
from ..general import (
    OVALEntity,
    load_property_and_notes_of_oval_entity,
    required_attribute,
)


def load_variable(oval_variable_xml_el):
    variable_property, notes = load_property_and_notes_of_oval_entity(
        oval_variable_xml_el
    )

    variable = Variable(
        oval_variable_xml_el.tag,
        required_attribute(oval_variable_xml_el, "id"),
        required_attribute(oval_variable_xml_el, "datatype"),
        variable_property,
    )
    variable.comment = oval_variable_xml_el.get("comment", "")
    variable.deprecated = STR_TO_BOOL.get(
        oval_variable_xml_el.get("deprecated", ""), False
    )
    variable.notes = notes
    variable.version = required_attribute(oval_variable_xml_el, "version")
    return variable


class Variable(OVALEntity):
    def __init__(self, tag, id_, data_type, properties):
        super(Variable, self).__init__(tag, id_, properties)
        self.data_type = data_type

    def get_xml_element(self):
        return super(Variable, self).get_xml_element(datatype=self.data_type)

    def get_variable_references(self):
        return self._get_references("var_ref")

    def get_object_references(self):
        return self._get_references("object_ref")
